cimport cython
import numpy as np
cimport numpy as np
import sys

from yt.geometry.selection_routines cimport \
    SelectorObject, AlwaysSelector, OctreeSubsetSelector
from yt.utilities.lib.fp_utils cimport imax
from yt.geometry.oct_container cimport \
    SparseOctreeContainer, OctObjectPool
from yt.geometry.oct_visitors cimport Oct
from yt.geometry.particle_deposit cimport \
    ParticleDepositOperation
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy
import data_structures
from yt.utilities.lib.misc_utilities import OnceIndirect

cdef extern from "platform_dep.h":
    ctypedef int int32_t
    ctypedef long long int64_t
    void *alloca(int)

cdef extern from "cosmology.h":
    ctypedef struct CosmologyParameters "CosmologyParameters" :
        pass
    CosmologyParameters *cosmology_allocate()
    void cosmology_free(CosmologyParameters *c)
    void cosmology_set_fixed(CosmologyParameters *c)

    void cosmology_set_OmegaM(CosmologyParameters *c, double value)
    void cosmology_set_OmegaB(CosmologyParameters *c, double value)
    void cosmology_set_OmegaL(CosmologyParameters *c, double value)
    void cosmology_set_h(CosmologyParameters *c, double value)
    void cosmology_set_DeltaDC(CosmologyParameters *c, double value)

    double abox_from_auni(CosmologyParameters *c, double a) nogil
    double tcode_from_auni(CosmologyParameters *c, double a) nogil
    double tphys_from_auni(CosmologyParameters *c, double a) nogil

    double auni_from_abox(CosmologyParameters *c, double v) nogil
    double auni_from_tcode(CosmologyParameters *c, double v) nogil
    double auni_from_tphys(CosmologyParameters *c, double v) nogil

    double abox_from_tcode(CosmologyParameters *c, double tcode) nogil
    double tcode_from_abox(CosmologyParameters *c, double abox) nogil

    double tphys_from_abox(CosmologyParameters *c, double abox) nogil
    double tphys_from_tcode(CosmologyParameters *c, double tcode) nogil

cdef extern from "artio.h":
    ctypedef struct artio_fileset_handle "artio_fileset" :
        pass
    ctypedef struct artio_selection "artio_selection" :
        pass
    ctypedef struct artio_context :
        pass
    cdef extern artio_context *artio_context_global

    # open modes
    cdef int ARTIO_OPEN_HEADER "ARTIO_OPEN_HEADER"
    cdef int ARTIO_OPEN_GRID "ARTIO_OPEN_GRID"
    cdef int ARTIO_OPEN_PARTICLES "ARTIO_OPEN_PARTICLES"

    # parameter constants
    cdef int ARTIO_TYPE_STRING "ARTIO_TYPE_STRING"
    cdef int ARTIO_TYPE_CHAR "ARTIO_TYPE_CHAR"
    cdef int ARTIO_TYPE_INT "ARTIO_TYPE_INT"
    cdef int ARTIO_TYPE_FLOAT "ARTIO_TYPE_FLOAT"
    cdef int ARTIO_TYPE_DOUBLE "ARTIO_TYPE_DOUBLE"
    cdef int ARTIO_TYPE_LONG "ARTIO_TYPE_LONG"

    cdef int ARTIO_MAX_STRING_LENGTH "ARTIO_MAX_STRING_LENGTH"

    cdef int ARTIO_PARAMETER_EXHAUSTED "ARTIO_PARAMETER_EXHAUSTED"

    # grid read options
    cdef int ARTIO_READ_LEAFS "ARTIO_READ_LEAFS"
    cdef int ARTIO_READ_REFINED "ARTIO_READ_REFINED"
    cdef int ARTIO_READ_ALL "ARTIO_READ_ALL"
    cdef int ARTIO_READ_REFINED_NOT_ROOT "ARTIO_READ_REFINED_NOT_ROOT"
    cdef int ARTIO_RETURN_CELLS "ARTIO_RETURN_CELLS"
    cdef int ARTIO_RETURN_OCTS "ARTIO_RETURN_OCTS"

    # errors
    cdef int ARTIO_SUCCESS "ARTIO_SUCCESS"
    cdef int ARTIO_ERR_MEMORY_ALLOCATION "ARTIO_ERR_MEMORY_ALLOCATION"

    artio_fileset_handle *artio_fileset_open(char *file_prefix, int type, artio_context *context )
    int artio_fileset_close( artio_fileset_handle *handle )
    int artio_fileset_open_particle( artio_fileset_handle *handle )
    int artio_fileset_open_grid(artio_fileset_handle *handle)
    int artio_fileset_close_grid(artio_fileset_handle *handle)

    int artio_fileset_has_grid( artio_fileset_handle *handle )
    int artio_fileset_has_particles( artio_fileset_handle *handle )

    # selection functions
    artio_selection *artio_selection_allocate( artio_fileset_handle *handle )
    artio_selection *artio_select_all( artio_fileset_handle *handle )
    artio_selection *artio_select_volume( artio_fileset_handle *handle, double lpos[3], double rpos[3] )
    int artio_selection_add_root_cell( artio_selection *selection, int coords[3] )
    int artio_selection_destroy( artio_selection *selection )
    int artio_selection_iterator( artio_selection *selection,
            int64_t max_range_size, int64_t *start, int64_t *end )
    int64_t artio_selection_size( artio_selection *selection )
    void artio_selection_print( artio_selection *selection )

    # parameter functions
    int artio_parameter_iterate( artio_fileset_handle *handle, char *key, int *type, int *length )
    int artio_parameter_get_int_array(artio_fileset_handle *handle, char * key, int length, int32_t *values)
    int artio_parameter_get_float_array(artio_fileset_handle *handle, char * key, int length, float *values)
    int artio_parameter_get_long_array(artio_fileset_handle *handle, char * key, int length, int64_t *values)
    int artio_parameter_get_double_array(artio_fileset_handle *handle, char * key, int length, double *values)
    int artio_parameter_get_string_array(artio_fileset_handle *handle, char * key, int length, char **values )

    # grid functions
    int artio_grid_cache_sfc_range(artio_fileset_handle *handle, int64_t start, int64_t end)
    int artio_grid_clear_sfc_cache( artio_fileset_handle *handle )

    int artio_grid_read_root_cell_begin(artio_fileset_handle *handle, int64_t sfc,
        double *pos, float *variables,
        int *num_tree_levels, int *num_octs_per_level)
    int artio_grid_read_root_cell_end(artio_fileset_handle *handle)

    int artio_grid_read_level_begin(artio_fileset_handle *handle, int level )
    int artio_grid_read_level_end(artio_fileset_handle *handle)

    int artio_grid_read_oct(artio_fileset_handle *handle, double *pos,
            float *variables, int *refined)

    int artio_grid_count_octs_in_sfc_range(artio_fileset_handle *handle,
            int64_t start, int64_t end, int64_t *num_octs)

    #particle functions
    int artio_fileset_open_particles(artio_fileset_handle *handle)
    int artio_particle_read_root_cell_begin(artio_fileset_handle *handle, int64_t sfc,
                        int * num_particle_per_species)
    int artio_particle_read_root_cell_end(artio_fileset_handle *handle)
    int artio_particle_read_particle(artio_fileset_handle *handle, int64_t *pid, int *subspecies,
                        double *primary_variables, float *secondary_variables)
    int artio_particle_cache_sfc_range(artio_fileset_handle *handle, int64_t sfc_start, int64_t sfc_end)
    int artio_particle_clear_sfc_cache(artio_fileset_handle *handle)
    int artio_particle_read_species_begin(artio_fileset_handle *handle, int species)
    int artio_particle_read_species_end(artio_fileset_handle *handle)


cdef extern from "artio_internal.h":
    np.int64_t artio_sfc_index( artio_fileset_handle *handle, int coords[3] ) nogil
    void artio_sfc_coords( artio_fileset_handle *handle, int64_t index, int coords[3] ) nogil

cdef void check_artio_status(int status, char *fname="[unknown]"):
    if status != ARTIO_SUCCESS:
        import traceback
        traceback.print_stack()
        callername = sys._getframe().f_code.co_name
        nline = sys._getframe().f_lineno
        raise RuntimeError('failure with status', status, 'in function',fname,'from caller', callername, nline)

cdef class artio_fileset :
    cdef public object parameters
    cdef artio_fileset_handle *handle

    # cosmology parameter for time unit conversion
    cdef CosmologyParameters *cosmology
    cdef float tcode_to_years

    # common attributes
    cdef public int num_grid
    cdef int64_t num_root_cells
    cdef int64_t sfc_min, sfc_max

    # grid attributes
    cdef public int has_grid
    cdef public int min_level, max_level
    cdef public int num_grid_variables
    cdef int *num_octs_per_level
    cdef float *grid_variables

    # particle attributes
    cdef public int has_particles
    cdef public int num_species
    cdef int *particle_position_index
    cdef int *num_particles_per_species
    cdef double *primary_variables
    cdef float *secondary_variables

    def __init__(self, char *file_prefix) :
        cdef int artio_type = ARTIO_OPEN_HEADER
        cdef int64_t num_root

        self.handle = artio_fileset_open( file_prefix, artio_type, artio_context_global )
        if not self.handle :
            raise RuntimeError

        self.read_parameters()

        self.num_root_cells = self.parameters['num_root_cells'][0]
        self.num_grid = 1
        num_root = self.num_root_cells
        while num_root > 1 :
            self.num_grid <<= 1
            num_root >>= 3

        self.sfc_min = 0
        self.sfc_max = self.num_root_cells-1

        # initialize cosmology module
        if "abox" in self.parameters:
            self.cosmology = cosmology_allocate()
            cosmology_set_OmegaM(self.cosmology, self.parameters['OmegaM'][0])
            cosmology_set_OmegaL(self.cosmology, self.parameters['OmegaL'][0])
            cosmology_set_OmegaB(self.cosmology, self.parameters['OmegaB'][0])
            cosmology_set_h(self.cosmology, self.parameters['hubble'][0])
            cosmology_set_DeltaDC(self.cosmology, self.parameters['DeltaDC'][0])
            cosmology_set_fixed(self.cosmology)
        else:
            self.cosmology = NULL
            self.tcode_to_years = self.parameters['time_unit'][0]/(365.25*86400)

        # grid detection
        self.min_level = 0
        self.max_level = self.parameters['grid_max_level'][0]
        self.num_grid_variables = self.parameters['num_grid_variables'][0]

        self.num_octs_per_level = <int *>malloc(self.max_level*sizeof(int))
        self.grid_variables = <float *>malloc(8*self.num_grid_variables*sizeof(float))
        if (not self.num_octs_per_level) or (not self.grid_variables) :
            raise MemoryError

        if artio_fileset_has_grid(self.handle):
            status = artio_fileset_open_grid(self.handle)
            check_artio_status(status)
            self.has_grid = 1
        else:
            self.has_grid = 0

        # particle detection
        if ( artio_fileset_has_particles(self.handle) ):
            status = artio_fileset_open_particles(self.handle)
            check_artio_status(status)
            self.has_particles = 1
	    
            for v in ["num_particle_species","num_primary_variables","num_secondary_variables"]:
                if v not in self.parameters:
                    raise RuntimeError("Unable to locate particle header information in artio header: key=", v)

            self.num_species = self.parameters['num_particle_species'][0]
            self.particle_position_index = <int *>malloc(3*sizeof(int)*self.num_species)
            if not self.particle_position_index :
                raise MemoryError
            for ispec in range(self.num_species) :
                species_labels = "species_%02d_primary_variable_labels"% (ispec,)
                if species_labels not in self.parameters:
                    raise RuntimeError("Unable to locate variable labels for species",ispec)

                labels = self.parameters[species_labels]
                try :
                    self.particle_position_index[3*ispec+0] = labels.index('POSITION_X')
                    self.particle_position_index[3*ispec+1] = labels.index('POSITION_Y')
                    self.particle_position_index[3*ispec+2] = labels.index('POSITION_Z')
                except ValueError :
                    raise RuntimeError("Unable to locate position information for particle species", ispec)

            self.num_particles_per_species =  <int *>malloc(sizeof(int)*self.num_species)
            self.primary_variables = <double *>malloc(sizeof(double)*max(self.parameters['num_primary_variables']))
            self.secondary_variables = <float *>malloc(sizeof(float)*max(self.parameters['num_secondary_variables']))
            if (not self.num_particles_per_species) or (not self.primary_variables) or (not self.secondary_variables) :
                raise MemoryError
        else:
            self.has_particles = 0

    def __dealloc__(self) :
        if self.num_octs_per_level : free(self.num_octs_per_level)
        if self.grid_variables : free(self.grid_variables)

        if self.particle_position_index : free(self.particle_position_index)
        if self.num_particles_per_species : free(self.num_particles_per_species)
        if self.primary_variables : free(self.primary_variables)
        if self.secondary_variables : free(self.secondary_variables)

        if self.cosmology : cosmology_free(self.cosmology)

        if self.handle : artio_fileset_close(self.handle)

    def read_parameters(self) :
        from sys import version
        cdef char key[64]
        cdef int type
        cdef int length
        cdef char ** char_values
        cdef int32_t *int_values
        cdef int64_t *long_values
        cdef float *float_values
        cdef double *double_values

        self.parameters = {}

        while artio_parameter_iterate( self.handle, key, &type, &length ) == ARTIO_SUCCESS :
	    
            if type == ARTIO_TYPE_STRING :
                char_values = <char **>malloc(length*sizeof(char *))
                for i in range(length) :
                    char_values[i] = <char *>malloc( ARTIO_MAX_STRING_LENGTH*sizeof(char) )
                artio_parameter_get_string_array( self.handle, key, length, char_values )
                parameter = [ char_values[i] for i in range(length) ]
                for i in range(length) :
                    free(char_values[i])
                free(char_values)
                if version[0] == '3':
                    for i in range(0,len(parameter)):
                        parameter[i] = parameter[i].decode('utf-8')
            elif type == ARTIO_TYPE_INT :
                int_values = <int32_t *>malloc(length*sizeof(int32_t))
                artio_parameter_get_int_array( self.handle, key, length, int_values )
                parameter = [ int_values[i] for i in range(length) ]
                free(int_values)
            elif type == ARTIO_TYPE_LONG :
                long_values = <int64_t *>malloc(length*sizeof(int64_t))
                artio_parameter_get_long_array( self.handle, key, length, long_values )
                parameter = [ long_values[i] for i in range(length) ]
                free(long_values)
            elif type == ARTIO_TYPE_FLOAT :
                float_values = <float *>malloc(length*sizeof(float))
                artio_parameter_get_float_array( self.handle, key, length, float_values )
                parameter = [ float_values[i] for i in range(length) ]
                free(float_values)
            elif type == ARTIO_TYPE_DOUBLE :
                double_values = <double *>malloc(length*sizeof(double))
                artio_parameter_get_double_array( self.handle, key, length, double_values )
                parameter = [ double_values[i] for i in range(length) ]
                free(double_values)
            else :
                raise RuntimeError("ARTIO file corruption detected: invalid type!")

            if version[0] == '3':
                self.parameters[key.decode('utf-8')] = parameter
            else:
                self.parameters[key] = parameter

    def abox_from_auni(self, np.float64_t a):
        if self.cosmology:
            return abox_from_auni(self.cosmology, a)
        else:
            raise RuntimeError("abox_from_auni called for non-cosmological ARTIO fileset!")

    def tcode_from_auni(self, np.float64_t a):
        if self.cosmology:
            return tcode_from_auni(self.cosmology, a)
        else:
            raise RuntimeError("tcode_from_auni called for non-cosmological ARTIO fileset!")

    def tphys_from_auni(self, np.float64_t a):
        if self.cosmology:
            return tphys_from_auni(self.cosmology, a)
        else:
            raise RuntimeError("tphys_from_auni called for non-cosmological ARTIO fileset!")

    def auni_from_abox(self, np.float64_t v):
        if self.cosmology:
            return auni_from_abox(self.cosmology, v)
        else:
            raise RuntimeError("auni_from_abox called for non-cosmological ARTIO fileset!")

    def auni_from_tcode(self, np.float64_t v):
        if self.cosmology:
            return auni_from_tcode(self.cosmology, v)
        else:
            raise RuntimeError("auni_from_tcode called for non-cosmological ARTIO fileset!")

    @cython.wraparound(False)
    @cython.boundscheck(False)
    def auni_from_tcode_array(self, np.ndarray[np.float64_t] tcode):
        cdef int i, nauni
        cdef np.ndarray[np.float64_t, ndim=1] auni
        cdef CosmologyParameters *cosmology = self.cosmology
        if not cosmology:
            raise RuntimeError("auni_from_tcode_array called for non-cosmological ARTIO fileset!")
        auni = np.empty_like(tcode)
        nauni = auni.shape[0]
        with nogil:
            for i in range(nauni):
                auni[i] = auni_from_tcode(self.cosmology, tcode[i])
        return auni

    def auni_from_tphys(self, np.float64_t v):
        if self.cosmology:
            return auni_from_tphys(self.cosmology, v)
        else:
            raise RuntimeError("auni_from_tphys called for non-cosmological ARTIO fileset!")

    def abox_from_tcode(self, np.float64_t abox):
        if self.cosmology:
            return abox_from_tcode(self.cosmology, abox)
        else:
            raise RuntimeError("abox_from_tcode called for non-cosmological ARTIO fileset!")

    def tcode_from_abox(self, np.float64_t abox):
        if self.cosmology:
            return tcode_from_abox(self.cosmology, abox)
        else:
            raise RuntimeError("tcode_from_abox called for non-cosmological ARTIO fileset!")

    def tphys_from_abox(self, np.float64_t abox):
        if self.cosmology:
            return tphys_from_abox(self.cosmology, abox)
        else:
            raise RuntimeError("tphys_from_abox called for non-cosmological ARTIO fileset!")

    def tphys_from_tcode(self, np.float64_t tcode):
        if self.cosmology:
            return tphys_from_tcode(self.cosmology, tcode)
        else:
            return self.tcode_to_years*tcode

    @cython.wraparound(False)
    @cython.boundscheck(False)
    def tphys_from_tcode_array(self, np.ndarray[np.float64_t] tcode):
        cdef int i, ntphys
        cdef np.ndarray[np.float64_t, ndim=1] tphys
        cdef CosmologyParameters *cosmology = self.cosmology
        tphys = np.empty_like(tcode)
        ntphys = tcode.shape[0]

        if cosmology:
            tphys = np.empty_like(tcode)
            ntphys = tcode.shape[0]
            with nogil:
                for i in range(ntphys):
                    tphys[i] = tphys_from_tcode(cosmology, tcode[i])
            return tphys
        else:
            return tcode*self.tcode_to_years

#    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def read_particle_chunk(self, SelectorObject selector, int64_t sfc_start, int64_t sfc_end, fields) :
        cdef int i
        cdef int status
        cdef int subspecies
        cdef int64_t pid

        cdef int num_fields = len(fields)
        cdef np.float64_t pos[3]

        # since RuntimeErrors are not fatal, ensure no artio_particles* functions
        # called if fileset lacks particles
        if not self.has_particles: return

        data = {}
        accessed_species = np.zeros( self.num_species, dtype="int")
        selected_mass = [ None for i in range(self.num_species)]
        selected_pid = [ None for i in range(self.num_species)]
        selected_species = [ None for i in range(self.num_species)]
        selected_primary = [ [] for i in range(self.num_species)]
        selected_secondary = [ [] for i in range(self.num_species)]

        for species,field in fields :
            if species < 0 or species > self.num_species :
                raise RuntimeError("Invalid species provided to read_particle_chunk")
            accessed_species[species] = 1

            if self.parameters["num_primary_variables"][species] > 0 and \
                    field in self.parameters["species_%02u_primary_variable_labels"%(species,)] :
                selected_primary[species].append((self.parameters["species_%02u_primary_variable_labels"%(species,)].index(field),(species,field)))
                data[(species,field)] = np.empty(0,dtype="float64")
            elif self.parameters["num_secondary_variables"][species] > 0 and \
                    field in self.parameters["species_%02u_secondary_variable_labels"%(species,)] :
                selected_secondary[species].append((self.parameters["species_%02u_secondary_variable_labels"%(species,)].index(field),(species,field)))
                data[(species,field)] = np.empty(0,dtype="float64")
            elif field == "MASS" :
                selected_mass[species] = (species,field)
                data[(species,field)] = np.empty(0,dtype="float64")
            elif field == "PID" :
                selected_pid[species] = (species,field)
                data[(species,field)] = np.empty(0,dtype="int64")
            elif field == "SPECIES" :
                selected_species[species] = (species,field)
                data[(species,field)] = np.empty(0,dtype="int8")
            else :
                raise RuntimeError("invalid field name provided to read_particle_chunk")

        # cache the range
        status = artio_particle_cache_sfc_range( self.handle, self.sfc_min, self.sfc_max )
        check_artio_status(status)

        for sfc in range( sfc_start, sfc_end+1 ) :
            status = artio_particle_read_root_cell_begin( self.handle, sfc,
                    self.num_particles_per_species )
            check_artio_status(status)

            for ispec in range(self.num_species) :
                if accessed_species[ispec] :
                    status = artio_particle_read_species_begin(self.handle, ispec)
                    check_artio_status(status)

                    for particle in range( self.num_particles_per_species[ispec] ) :
                        status = artio_particle_read_particle(self.handle,
                                &pid, &subspecies, self.primary_variables,
                                self.secondary_variables)
                        check_artio_status(status)

                        for i in range(3) :
                            pos[i] = self.primary_variables[self.particle_position_index[3*ispec+i]]

                        if selector.select_point(pos) :
                            # loop over primary variables
                            for i,field in selected_primary[ispec] :
                                count = len(data[field])
                                data[field].resize(count+1)
                                data[field][count] = self.primary_variables[i]

                            # loop over secondary variables
                            for i,field in selected_secondary[ispec] :
                                count = len(data[field])
                                data[field].resize(count+1)
                                data[field][count] = self.secondary_variables[i]

                            # add particle id
                            if selected_pid[ispec] :
                                count = len(data[selected_pid[ispec]])
                                data[selected_pid[ispec]].resize(count+1)
                                data[selected_pid[ispec]][count] = pid

                            # add mass if requested
                            if selected_mass[ispec] :
                                count = len(data[selected_mass[ispec]])
                                data[selected_mass[ispec]].resize(count+1)
                                data[selected_mass[ispec]][count] = self.parameters["particle_species_mass"][ispec]

                    status = artio_particle_read_species_end( self.handle )
                    check_artio_status(status)

            status = artio_particle_read_root_cell_end( self.handle )
            check_artio_status(status)

        return data

    #@cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def read_grid_chunk(self, SelectorObject selector, int64_t sfc_start, int64_t sfc_end, fields ):
        cdef int i
        cdef int level
        cdef int num_oct_levels
        cdef int refined[8]
        cdef int status
        cdef int64_t count
        cdef int64_t max_octs
        cdef double dpos[3]
        cdef np.float64_t left[3]
        cdef np.float64_t right[3]
        cdef np.float64_t dds[3]

        cdef int *field_order
        cdef int num_fields  = len(fields)
        field_order = <int*>malloc(sizeof(int)*num_fields)

        # translate fields from ARTIO names to indices
        var_labels = self.parameters['grid_variable_labels']
        for i, f in enumerate(fields):
            if f not in var_labels:
                raise RuntimeError("Field",f,"is not known to ARTIO")
            field_order[i] = var_labels.index(f)

        status = artio_grid_cache_sfc_range( self.handle, self.sfc_min, self.sfc_max )
        check_artio_status(status)

        # determine max number of cells we could hit (optimize later)
        #status = artio_grid_count_octs_in_sfc_range( self.handle,
        #        sfc_start, sfc_end, &max_octs )
        #check_artio_status(status)
        #max_cells = sfc_end-sfc_start+1 + max_octs*8

        # allocate space for _fcoords, _icoords, _fwidth, _ires
        #fcoords = np.empty((max_cells, 3), dtype="float64")
        #ires = np.empty(max_cells, dtype="int64")
        fcoords = np.empty((0, 3), dtype="float64")
        ires = np.empty(0, dtype="int64")

        #data = [ np.empty(max_cells, dtype="float32") for i in range(num_fields) ]
        data = [ np.empty(0,dtype="float64") for i in range(num_fields)]

        count = 0
        for sfc in range( sfc_start, sfc_end+1 ) :
            status = artio_grid_read_root_cell_begin( self.handle, sfc,
                    dpos, self.grid_variables, &num_oct_levels, self.num_octs_per_level )
            check_artio_status(status)

            if num_oct_levels == 0 :
                for i in range(num_fields) :
                    data[i].resize(count+1)
                    data[i][count] = self.grid_variables[field_order[i]]
                fcoords.resize((count+1,3))
                for i in range(3) :
                    fcoords[count][i] = dpos[i]
                ires.resize(count+1)
                ires[count] = 0
                count += 1

            for level in range(1,num_oct_levels+1) :
                status = artio_grid_read_level_begin( self.handle, level )
                check_artio_status(status)

                for i in range(3) :
                    dds[i] = 2.**-level

                for oct in range(self.num_octs_per_level[level-1]) :
                    status = artio_grid_read_oct( self.handle, dpos, self.grid_variables, refined )
                    check_artio_status(status)

                    for child in range(8) :
                        if not refined[child] :
                            for i in range(3) :
                                left[i] = (dpos[i]-dds[i]) if (child & (i<<1)) else dpos[i]
                                right[i] = left[i] + dds[i]

                            if selector.select_bbox(left,right) :
                                fcoords.resize((count+1, 3))
                                for i in range(3) :
                                    fcoords[count][i] = left[i]+0.5*dds[i]
                                ires.resize(count+1)
                                ires[count] = level
                                for i in range(num_fields) :
                                    data[i].resize(count+1)
                                    data[i][count] = self.grid_variables[self.num_grid_variables*child+field_order[i]]
                                count += 1
                status = artio_grid_read_level_end( self.handle )
                check_artio_status(status)

            status = artio_grid_read_root_cell_end( self.handle )
            check_artio_status(status)

        free(field_order)

        #fcoords.resize((count,3))
        #ires.resize(count)
        #
        #for i in range(num_fields) :
        #    data[i].resize(count)

        return (fcoords, ires, data)

    def root_sfc_ranges_all(self, int max_range_size = 1024) :
        cdef int64_t sfc_start, sfc_end
        cdef artio_selection *selection

        selection = artio_select_all( self.handle )
        if selection == NULL :
            raise RuntimeError
        sfc_ranges = []
        while artio_selection_iterator(selection, max_range_size,
                &sfc_start, &sfc_end) == ARTIO_SUCCESS :
            sfc_ranges.append([sfc_start, sfc_end])
        artio_selection_destroy(selection)
        return sfc_ranges

    def root_sfc_ranges(self, SelectorObject selector,
                        int max_range_size = 1024):
        cdef int coords[3]
        cdef int64_t sfc_start, sfc_end
        cdef np.float64_t left[3]
        cdef np.float64_t right[3]
        cdef np.float64_t dds[3]
        cdef artio_selection *selection
        cdef int i, j, k

        sfc_ranges=[]
        selection = artio_selection_allocate(self.handle)
        for i in range(self.num_grid) :
            # stupid cython
            coords[0] = i
            left[0] = coords[0]
            right[0] = left[0] + 1.0
            for j in range(self.num_grid) :
                coords[1] = j
                left[1] = coords[1]
                right[1] = left[1] + 1.0
                for k in range(self.num_grid) :
                    coords[2] = k
                    left[2] = coords[2]
                    right[2] = left[2] + 1.0
                    if selector.select_bbox(left,right) :
                        status = artio_selection_add_root_cell(selection, coords)
                        check_artio_status(status)

        while artio_selection_iterator(selection, max_range_size,
                &sfc_start, &sfc_end) == ARTIO_SUCCESS :
            sfc_ranges.append([sfc_start, sfc_end])

        artio_selection_destroy(selection)
        return sfc_ranges

###################################################
def artio_is_valid( char *file_prefix ) :
    cdef artio_fileset_handle *handle = artio_fileset_open( file_prefix,
            ARTIO_OPEN_HEADER, artio_context_global )
    if handle == NULL :
        return False
    else :
        artio_fileset_close(handle)
    return True

cdef class ARTIOSFCRangeHandler:
    cdef public np.int64_t sfc_start
    cdef public np.int64_t sfc_end
    cdef public artio_fileset artio_handle
    cdef public object root_mesh_handler
    cdef public object oct_count
    cdef public object octree_handler
    cdef artio_fileset_handle *handle
    cdef np.float64_t DLE[3]
    cdef np.float64_t DRE[3]
    cdef np.float64_t dds[3]
    cdef np.int64_t dims[3]
    cdef public np.int64_t total_octs
    cdef np.int64_t *doct_count
    cdef np.int64_t **pcount
    cdef float **root_mesh_data
    cdef np.int64_t nvars[2]
    cdef int cache_root_mesh

    def __init__(self, domain_dimensions, # cells
                 domain_left_edge,
                 domain_right_edge,
                 artio_fileset artio_handle,
                 sfc_start, sfc_end, int cache_root_mesh = 0):
        cdef int i
        cdef np.int64_t sfc
        self.sfc_start = sfc_start
        self.sfc_end = sfc_end
        self.artio_handle = artio_handle
        self.root_mesh_handler = None
        self.octree_handler = None
        self.handle = artio_handle.handle
        self.oct_count = None
        self.root_mesh_data = NULL
        self.pcount = NULL
        self.cache_root_mesh = cache_root_mesh

        if artio_handle.has_particles:
            self.pcount = <np.int64_t **> malloc(sizeof(np.int64_t*)
                * artio_handle.num_species)
            self.nvars[0] = artio_handle.num_species
            for i in range(artio_handle.num_species):
                self.pcount[i] = <np.int64_t*> malloc(sizeof(np.int64_t)
                    * (self.sfc_end - self.sfc_start + 1))
                for sfc in range(self.sfc_end - self.sfc_start + 1):
                    self.pcount[i][sfc] = 0
        else:
            self.nvars[0] = 0

        if artio_handle.has_grid:
            self.nvars[1] = artio_handle.num_grid_variables
        else:
            self.nvars[1] = 0

        for i in range(3):
            self.dims[i] = domain_dimensions[i]
            self.DLE[i] = domain_left_edge[i]
            self.DRE[i] = domain_right_edge[i]
            self.dds[i] = (self.DRE[i] - self.DLE[i])/self.dims[i]

    def __dealloc__(self):
        cdef int i
        if self.artio_handle.has_particles:
            for i in range(self.nvars[0]):
                free(self.pcount[i])
            free(self.pcount)
        if self.artio_handle.has_grid:
            if self.root_mesh_data != NULL:
                for i in range(self.nvars[1]):
                    free(self.root_mesh_data[i])
                free(self.root_mesh_data)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def construct_mesh(self):
        cdef int status, level, ngv
        cdef np.int64_t sfc, oc, i
        cdef double dpos[3]
        cdef int num_oct_levels
        cdef int max_level = self.artio_handle.max_level
        cdef int *num_octs_per_level = <int *>malloc(
            (max_level + 1)*sizeof(int))
        cdef int num_species = self.artio_handle.num_species
        cdef int *num_particles_per_species
        cdef ARTIOOctreeContainer octree
        ngv = self.nvars[1]
        cdef float *grid_variables = <float *>malloc(
            ngv * sizeof(float))
        self.octree_handler = octree = ARTIOOctreeContainer(self)
        if self.cache_root_mesh == 1:
            self.root_mesh_data = <float **>malloc(sizeof(float *) * ngv)
            for i in range(ngv):
                self.root_mesh_data[i] = <float *>malloc(sizeof(float) * \
                    (self.sfc_end - self.sfc_start + 1))
        # We want to pre-allocate an array of root pointers.  In the future,
        # this will be pre-determined by the ARTIO library.  However, because
        # realloc plays havoc with our tree searching, we can't utilize an
        # expanding array at the present time.
        octree.allocate_domains([], self.sfc_end - self.sfc_start + 1)
        cdef np.ndarray[np.int64_t, ndim=1] oct_count
        oct_count = np.zeros(self.sfc_end - self.sfc_start + 1, dtype="int64")
        status = artio_grid_cache_sfc_range(self.handle, self.sfc_start,
                                            self.sfc_end)
        check_artio_status(status)
        for sfc in range(self.sfc_start, self.sfc_end + 1):
            status = artio_grid_read_root_cell_begin( self.handle,
                sfc, dpos, grid_variables, &num_oct_levels,
                num_octs_per_level)
            check_artio_status(status)
            for i in range(ngv * self.cache_root_mesh):
                self.root_mesh_data[i][sfc - self.sfc_start] = \
                    grid_variables[i]
            if num_oct_levels > 0:
                oc = 0
                for level in range(num_oct_levels):
                    oc += num_octs_per_level[level]
                self.total_octs += oc
                oct_count[sfc - self.sfc_start] = oc
                octree.initialize_local_mesh(oc, num_oct_levels,
                    num_octs_per_level, sfc)
            status = artio_grid_read_root_cell_end(self.handle)
            check_artio_status(status)
        status = artio_grid_clear_sfc_cache(self.handle)
        check_artio_status(status)
        if self.artio_handle.has_particles:
            num_particles_per_species =  <int *>malloc(
                    sizeof(int)*num_species)

            # Now particles
            status = artio_particle_cache_sfc_range(self.handle, self.sfc_start,
                                            self.sfc_end)
            check_artio_status(status)
            for sfc in range(self.sfc_start, self.sfc_end + 1):
                # Now particles
                status = artio_particle_read_root_cell_begin(self.handle,
                        sfc, num_particles_per_species)
                check_artio_status(status)

                for i in range(num_species):
                    self.pcount[i][sfc - self.sfc_start] = \
                        num_particles_per_species[i]

                status = artio_particle_read_root_cell_end(self.handle)
                check_artio_status(status)

            status = artio_particle_clear_sfc_cache(self.handle)
            check_artio_status(status)

            free(num_particles_per_species)

        free(grid_variables)
        free(num_octs_per_level)
        self.oct_count = oct_count
        self.doct_count = <np.int64_t *> oct_count.data
        self.root_mesh_handler = ARTIORootMeshContainer(self)

    def free_mesh(self):
        self.octree_handler = None
        self.root_mesh_handler = None
        self.doct_count = NULL
        self.oct_count = None

def get_coords(artio_fileset handle, np.int64_t s):
    cdef int coords[3]
    artio_sfc_coords(handle.handle, s, coords)
    return (coords[0], coords[1], coords[2])

cdef struct particle_var_pointers:
    # The number of particles we have filled
    np.int64_t count
    # Number of primary variables and pointers to their indices
    int n_p
    int p_ind[16] # Max of 16 vars
    # Number of secondary variables and pointers to their indices
    int n_s
    int s_ind[16] # Max of 16 vars
    # Pointers to the bools and data arrays for mass, pid and species
    int n_mass
    np.float64_t *mass
    int n_pid
    np.int64_t *pid
    int n_species
    np.int8_t *species
    # Pointers to the pointers to primary and secondary vars
    np.float64_t *pvars[16]
    np.float64_t *svars[16]

cdef class ARTIOOctreeContainer(SparseOctreeContainer):
    # This is a transitory, created-on-demand OctreeContainer.  It should not
    # be considered to be long-lasting, and during its creation it will read
    # the index file.  This means that when created it will then be able to
    # provide coordinates, but then on subsequent IO accesses it will pass over
    # the file again, despite knowing the indexing system already.  Because of
    # this, we will avoid creating it as long as possible.

    cdef public artio_fileset artio_handle
    cdef ARTIOSFCRangeHandler range_handler
    cdef np.int64_t level_indices[32]
    cdef np.int64_t sfc_start
    cdef np.int64_t sfc_end

    def __init__(self, ARTIOSFCRangeHandler range_handler):
        self.range_handler = range_handler
        self.sfc_start = range_handler.sfc_start
        self.sfc_end = range_handler.sfc_end
        self.artio_handle = range_handler.artio_handle
        # Note the final argument is partial_coverage, which indicates whether
        # or not an Oct can be partially refined.
        dims, DLE, DRE = [], [], []
        for i in range(32):
            self.level_indices[i] = 0
        for i in range(3):
            # range_handler has dims in cells, which is the same as the number
            # of possible octs.  This is because we have a forest of octrees.
            dims.append(range_handler.dims[i])
            DLE.append(range_handler.DLE[i])
            DRE.append(range_handler.DRE[i])
        super(ARTIOOctreeContainer, self).__init__(dims, DLE, DRE)
        self.artio_handle = range_handler.artio_handle
        self.level_offset = 1
        self.domains = OctObjectPool()
        self.root_nodes = NULL

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void initialize_local_mesh(self, np.int64_t oct_count,
                              int num_oct_levels, int *num_octs_per_level,
                              np.int64_t sfc):
        # We actually will not be initializing the root mesh here, we will be
        # initializing the entire mesh between sfc_start and sfc_end.
        cdef np.int64_t oct_ind, tot_octs, ipos, nadded
        cdef int i, status, level, num_root, num_octs
        cdef int num_level_octs
        cdef artio_fileset_handle *handle = self.artio_handle.handle
        cdef int coords[3]
        cdef int max_level = self.artio_handle.max_level
        cdef double dpos[3]
        cdef np.float64_t f64pos[3]
        cdef np.float64_t dds[3]
        cdef np.ndarray[np.float64_t, ndim=2] pos
        # NOTE: We do not cache any SFC ranges here, as we should only ever be
        # called from within a pre-cached operation in the SFC handler.

        # We only allow one root oct.
        self.append_domain(oct_count)
        self.domains.containers[self.num_domains - 1].con_id = sfc

        oct_ind = -1
        ipos = 0
        for level in range(num_oct_levels):
            oct_ind = imax(oct_ind, num_octs_per_level[level])
            self.level_indices[level] = ipos
            ipos += num_octs_per_level[level]
        pos = np.empty((oct_ind, 3), dtype="float64")

        # Now we initialize
        # Note that we also assume we have already started reading the level.
        ipos = 0
        for level in range(num_oct_levels):
            status = artio_grid_read_level_begin(handle, level + 1)
            check_artio_status(status)
            for oct_ind in range(num_octs_per_level[level]):
                status = artio_grid_read_oct(handle, dpos, NULL, NULL)
                for i in range(3):
                    pos[oct_ind, i] = dpos[i]
                check_artio_status(status)
            status = artio_grid_read_level_end(handle)
            check_artio_status(status)
            nadded = self.add(self.num_domains, level, pos[:num_octs_per_level[level],:])
            if nadded != num_octs_per_level[level]:
                raise RuntimeError

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_sfc(self,
             np.ndarray[np.uint8_t, ndim=1] levels,
             np.ndarray[np.uint8_t, ndim=1] cell_inds,
             np.ndarray[np.int64_t, ndim=1] file_inds,
             np.ndarray[np.int64_t, ndim=1] domain_counts,
             field_indices, dest_fields):
        cdef np.ndarray[np.float32_t, ndim=2] source
        cdef np.ndarray[np.float64_t, ndim=1] dest
        cdef int n, status, i, di, num_oct_levels, nf, ngv, max_level
        cdef int level, j, oct_ind, si
        cdef np.int64_t sfc, ipos
        cdef np.float64_t val
        cdef artio_fileset_handle *handle = self.artio_handle.handle
        cdef double dpos[3]
        # We duplicate some of the grid_variables stuff here so that we can
        # potentially release the GIL
        nf = len(field_indices)
        ngv = self.artio_handle.num_grid_variables
        max_level = self.artio_handle.max_level
        cdef int *num_octs_per_level = <int *>malloc(
            (max_level + 1)*sizeof(int))
        cdef float *grid_variables = <float *>malloc(
            8 * ngv * sizeof(float))
        cdef int* field_ind = <int*> malloc(
            nf * sizeof(int))
        cdef np.float32_t **field_vals = <np.float32_t**> malloc(
            nf * sizeof(np.float32_t*))
        source_arrays = []
        ipos = -1
        for i in range(self.num_domains):
            ipos = imax(ipos, self.domains.containers[i].n)
        for i in range(nf):
            field_ind[i] = field_indices[i]
            # Note that we subtract one, because we're not using the root mesh.
            source = np.zeros((ipos, 8), dtype="float32")
            source_arrays.append(source)
            field_vals[i] = <np.float32_t*> source.data
        cdef np.int64_t level_position[32]
        cdef np.int64_t lp
        # First we need to walk the mesh in the file.  Then we fill in the dest
        # location based on the file index.
        # A few ways this could be improved:
        #   * Create a new visitor function that actually queried the data,
        #     rather than our somewhat hokey double-loop over SFC arrays.
        #   * Go to pointers for the dest arrays.
        #   * Enable preloading during mesh initialization
        #   * Calculate domain indices on the fly rather than with a
        #     double-loop to calculate domain_counts
        # The cons should be in order
        cdef np.int64_t sfc_start, sfc_end
        sfc_start = self.domains.containers[0].con_id
        sfc_end = self.domains.containers[self.num_domains - 1].con_id
        status = artio_grid_cache_sfc_range(handle, sfc_start, sfc_end)
        check_artio_status(status)
        cdef np.int64_t offset = 0
        for si in range(self.num_domains):
            sfc = self.domains.containers[si].con_id
            status = artio_grid_read_root_cell_begin( handle, sfc,
                    dpos, NULL, &num_oct_levels, num_octs_per_level)
            check_artio_status(status)
            lp = 0
            for level in range(num_oct_levels):
                status = artio_grid_read_level_begin(handle, level + 1)
                check_artio_status(status)
                level_position[level] = lp
                for oct_ind in range(num_octs_per_level[level]):
                    status = artio_grid_read_oct(handle, dpos, grid_variables, NULL)
                    check_artio_status(status)
                    for j in range(8):
                        for i in range(nf):
                            field_vals[i][(oct_ind + lp)*8+j] = \
                                grid_variables[ngv*j+field_ind[i]]
                status = artio_grid_read_level_end(handle)
                check_artio_status(status)
                lp += num_octs_per_level[level]
            status = artio_grid_read_root_cell_end(handle)
            check_artio_status(status)
            # Now we have all our sources.
            for j in range(nf):
                dest = dest_fields[j]
                source = source_arrays[j]
                for i in range(domain_counts[si]):
                    level = levels[i + offset]
                    oct_ind = file_inds[i + offset] + level_position[level]
                    dest[i + offset] = source[oct_ind, cell_inds[i + offset]]
            # Now, we offset by the actual number filled here.
            offset += domain_counts[si]
        status = artio_grid_clear_sfc_cache(handle)
        check_artio_status(status)
        free(field_ind)
        free(field_vals)
        free(grid_variables)
        free(num_octs_per_level)

    def fill_sfc_particles(self, fields):
        # This handles not getting particles for refined sfc values.
        rv = read_sfc_particles(self.artio_handle, self.sfc_start,
                                self.sfc_end, 0, fields,
                                self.range_handler.doct_count,
                                self.range_handler.pcount)
        return rv

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef read_sfc_particles(artio_fileset artio_handle,
                        np.int64_t sfc_start, np.int64_t sfc_end,
                        int read_unrefined, fields,
                        np.int64_t *doct_count,
                        np.int64_t **pcount):
    cdef int status, ispec, subspecies
    cdef np.int64_t sfc, particle, pid, ind, vind
    cdef int num_species = artio_handle.num_species
    cdef artio_fileset_handle *handle = artio_handle.handle
    cdef int num_oct_levels
    cdef int *num_particles_per_species =  <int *>malloc(
        sizeof(int)*num_species)
    cdef int *accessed_species =  <int *>malloc(
        sizeof(int)*num_species)
    cdef int *total_particles = <int *>malloc(
        sizeof(int)*num_species)
    cdef particle_var_pointers *vpoints = <particle_var_pointers *> malloc(
        sizeof(particle_var_pointers)*num_species)
    cdef double *primary_variables
    cdef double dpos[3]
    cdef float *secondary_variables
    cdef np.int64_t tp
    cdef int max_level = artio_handle.max_level
    cdef int *num_octs_per_level = <int *>malloc(
        (max_level + 1)*sizeof(int))

    cdef np.ndarray[np.int8_t, ndim=1] npi8arr
    cdef np.ndarray[np.int64_t, ndim=1] npi64arr
    cdef np.ndarray[np.float64_t, ndim=1] npf64arr

    if not artio_handle.has_particles:
        raise RuntimeError("Attempted to read non-existent particles in ARTIO")

    # Now we set up our field pointers
    params = artio_handle.parameters

    npri_vars = params["num_primary_variables"]
    nsec_vars = params["num_secondary_variables"]
    primary_variables = <double *>malloc(sizeof(double) * max(npri_vars))
    secondary_variables = <float *>malloc(sizeof(float) * max(nsec_vars))

    cdef particle_var_pointers *vp

    for ispec in range(num_species):
        total_particles[ispec] = 0
        accessed_species[ispec] = 0
        # Initialize our vpoints array
        vpoints[ispec].count = 0
        vpoints[ispec].n_mass = 0
        vpoints[ispec].n_pid = 0
        vpoints[ispec].n_species = 0
        vpoints[ispec].n_p = 0
        vpoints[ispec].n_s = 0

    # Pass through once.  We want every single particle.
    tp = 0
    cdef np.int64_t c
    for sfc in range(sfc_start, sfc_end + 1):
        c = doct_count[sfc - sfc_start]
        if read_unrefined == 1 and c > 0: continue
        if read_unrefined == 0 and c == 0: continue

        for ispec in range(num_species):
            total_particles[ispec] += pcount[ispec][sfc - sfc_start]

    # Now we allocate our final fields, which will be filled
    #for ispec in range(num_species):
    #    print "In SFC %s to %s reading %s of species %s" % (
    #        sfc_start, sfc_end + 1, total_particles[ispec], ispec)
    data = {}
    for species, field in sorted(fields):
        accessed_species[species] = 1
        pri_vars = params.get(
            "species_%02u_primary_variable_labels" % (species,), [])
        sec_vars = params.get(
            "species_%02u_secondary_variable_labels" % (species,), [])
        tp = total_particles[species]
        vp = &vpoints[species]
        if field == "PID":
            vp.n_pid = 1
            data[(species, field)] = np.zeros(tp, dtype="int64")
            npi64arr = data[(species, field)]
            vp.pid = <np.int64_t*> npi64arr.data
        elif field == "SPECIES":
            vp.n_species = 1
            data[(species, field)] = np.zeros(tp, dtype="int8")
            npi8arr = data[(species, field)]
            # We fill this *now*
            npi8arr += species
            vp.species = <np.int8_t*> npi8arr.data
        elif npri_vars[species] > 0 and field in pri_vars :
            data[(species, field)] = np.zeros(tp, dtype="float64")
            npf64arr = data[(species, field)]
            vp.p_ind[vp.n_p] = pri_vars.index(field)
            vp.pvars[vp.n_p] = <np.float64_t *> npf64arr.data
            vp.n_p += 1
        elif nsec_vars[species] > 0 and field in sec_vars :
            data[(species, field)] = np.zeros(tp, dtype="float64")
            npf64arr = data[(species, field)]
            vp.s_ind[vp.n_s] = sec_vars.index(field)
            vp.svars[vp.n_s] = <np.float64_t *> npf64arr.data
            vp.n_s += 1
        elif field == "MASS":
            vp.n_mass = 1
            data[(species, field)] = np.zeros(tp, dtype="float64")
            npf64arr = data[(species, field)]
            # We fill this *now*
            npf64arr += params["particle_species_mass"][species]
            vp.mass = <np.float64_t*> npf64arr.data

    status = artio_particle_cache_sfc_range( handle,
            sfc_start, sfc_end )
    check_artio_status(status)

    for sfc in range(sfc_start, sfc_end + 1):
        c = doct_count[sfc - sfc_start]
        check_artio_status(status)
        if read_unrefined == 1 and c > 0: continue
        if read_unrefined == 0 and c == 0: continue
        c = 0
        for ispec in range(num_species) :
            if accessed_species[ispec] == 0: continue
            c += pcount[ispec][sfc - sfc_start]
        if c == 0: continue
        status = artio_particle_read_root_cell_begin( handle, sfc,
                num_particles_per_species )
        check_artio_status(status)
        for ispec in range(num_species) :
            if accessed_species[ispec] == 0: continue
            status = artio_particle_read_species_begin(handle, ispec);
            check_artio_status(status)
            vp = &vpoints[ispec]

            for particle in range(num_particles_per_species[ispec]) :
                status = artio_particle_read_particle(handle,
                        &pid, &subspecies, primary_variables,
                        secondary_variables)
                check_artio_status(status)
                ind = vp.count

                for i in range(vp.n_p):
                    vind = vp.p_ind[i]
                    vp.pvars[i][ind] = primary_variables[vind]

                for i in range(vp.n_s):
                    vind = vp.s_ind[i]
                    vp.svars[i][ind] = secondary_variables[vind]

                if vp.n_pid:
                    vp.pid[ind] = pid

                vp.count += 1

            status = artio_particle_read_species_end( handle )
            check_artio_status(status)

        status = artio_particle_read_root_cell_end( handle )
        check_artio_status(status)

    status = artio_particle_clear_sfc_cache(handle)
    check_artio_status(status)

    free(num_octs_per_level)
    free(num_particles_per_species)
    free(total_particles)
    free(accessed_species)
    free(vpoints)
    free(primary_variables)
    free(secondary_variables)
    return data

cdef class ARTIORootMeshContainer:
    cdef public artio_fileset artio_handle
    cdef np.float64_t DLE[3]
    cdef np.float64_t DRE[3]
    cdef np.float64_t dds[3]
    cdef np.int64_t dims[3]
    cdef artio_fileset_handle *handle
    cdef np.uint64_t sfc_start
    cdef np.uint64_t sfc_end
    cdef public object _last_mask
    cdef public np.int64_t _last_selector_id
    cdef np.int64_t _last_mask_sum
    cdef ARTIOSFCRangeHandler range_handler
    cdef np.uint8_t *sfc_mask
    cdef np.int64_t nsfc

    def __init__(self, ARTIOSFCRangeHandler range_handler):
        cdef int i
        cdef np.int64_t sfci
        for i in range(3):
            self.DLE[i] = range_handler.DLE[i]
            self.DRE[i] = range_handler.DRE[i]
            self.dims[i] = range_handler.dims[i]
            self.dds[i] = range_handler.dds[i]
        self.handle = range_handler.handle
        self.artio_handle = range_handler.artio_handle
        self._last_mask = None
        self._last_selector_id = -1
        self.sfc_start = range_handler.sfc_start
        self.sfc_end = range_handler.sfc_end
        self.range_handler = range_handler
        # We assume that the number of octs has been created and filled
        # already.  We no longer care about ANY of the SFCs that have octs
        # inside them -- this goes for every operation that this object
        # performs.
        self.sfc_mask = <np.uint8_t *>malloc(sizeof(np.uint8_t) *
          self.sfc_end - self.sfc_start + 1)
        self.nsfc = 0
        for sfci in range(self.sfc_end - self.sfc_start + 1):
            if self.range_handler.oct_count[sfci] > 0:
                self.sfc_mask[sfci] = 0
            else:
                self.sfc_mask[sfci] = 1
                self.nsfc += 1

    def __dealloc__(self):
        free(self.sfc_mask)

    @cython.cdivision(True)
    cdef np.int64_t pos_to_sfc(self, np.float64_t pos[3]) nogil:
        # Calculate the index
        cdef int coords[3]
        cdef int i
        cdef np.int64_t sfc
        for i in range(3):
            coords[i] = <int>((pos[i] - self.DLE[i])/self.dds[i])
        sfc = artio_sfc_index(self.handle, coords)
        return sfc

    @cython.cdivision(True)
    cdef void sfc_to_pos(self, np.int64_t sfc, np.float64_t pos[3]) nogil:
        cdef int coords[3]
        cdef int i
        artio_sfc_coords(self.handle, sfc, coords)
        for i in range(3):
            pos[i] = self.DLE[i] + (coords[i] + 0.5) * self.dds[i]

    cdef np.int64_t count_cells(self, SelectorObject selector):
        # We visit each cell if it is not refined and determine whether it is
        # included or not.
        cdef np.int64_t sfc
        cdef np.float64_t pos[3]
        cdef np.float64_t right_edge[3]
        cdef int num_cells = 0
        cdef int i
        return self.mask(selector).sum()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def icoords(self, SelectorObject selector, np.int64_t num_cells = -1,
                int domain_id = -1):
        # Note that num_octs does not have to equal sfc_end - sfc_start + 1.
        cdef np.int64_t sfc, sfci = -1
        cdef int acoords[3]
        cdef int i
        cdef np.ndarray[np.uint8_t, ndim=1, cast=True] mask
        mask = self.mask(selector)
        num_cells = self._last_mask_sum
        cdef np.ndarray[np.int64_t, ndim=2] coords
        coords = np.empty((num_cells, 3), dtype="int64")
        cdef int filled = 0
        for sfc in range(self.sfc_start, self.sfc_end + 1):
            if self.sfc_mask[sfc - self.sfc_start] == 0: continue
            sfci += 1
            if mask[sfci] == 0: continue
            # Note that we do *no* checks on refinement here.  In fact, this
            # entire setup should not need to touch the disk except if the
            # artio sfc calculators need to.
            artio_sfc_coords(self.handle, sfc, acoords)
            for i in range(3):
                coords[filled, i] = acoords[i]
            filled += 1
        return coords

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fcoords(self, SelectorObject selector, np.int64_t num_cells = -1,
                int domain_id = -1):
        # Note that num_cells does not have to equal sfc_end - sfc_start + 1.
        cdef np.int64_t sfc, sfci = -1
        cdef np.float64_t pos[3]
        cdef int acoords[3]
        cdef int i
        cdef np.ndarray[np.uint8_t, ndim=1, cast=True] mask
        mask = self.mask(selector)
        num_cells = self._last_mask_sum
        cdef np.ndarray[np.float64_t, ndim=2] coords
        coords = np.empty((num_cells, 3), dtype="float64")
        cdef int filled = 0
        for sfc in range(self.sfc_start, self.sfc_end + 1):
            if self.sfc_mask[sfc - self.sfc_start] == 0: continue
            sfci += 1
            if mask[sfci] == 0: continue
            # Note that we do *no* checks on refinement here.  In fact, this
            # entire setup should not need to touch the disk except if the
            # artio sfc calculators need to.
            self.sfc_to_pos(sfc, pos)
            for i in range(3):
                coords[filled, i] = pos[i]
            filled += 1
        return coords

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fwidth(self, SelectorObject selector, np.int64_t num_cells = -1,
                int domain_id = -1):
        cdef int i
        cdef np.ndarray[np.uint8_t, ndim=1, cast=True] mask
        mask = self.mask(selector)
        num_cells = self._last_mask_sum
        cdef np.ndarray[np.float64_t, ndim=2] width
        width = np.zeros((num_cells, 3), dtype="float64")
        for i in range(3):
            width[:,i] = self.dds[i]
        return width

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def ires(self, SelectorObject selector, np.int64_t num_cells = -1,
                int domain_id = -1):
        cdef np.ndarray[np.uint8_t, ndim=1, cast=True] mask
        mask = self.mask(selector)
        num_cells = self._last_mask_sum
        cdef np.ndarray[np.int64_t, ndim=1] res
        res = np.zeros(num_cells, dtype="int64")
        return res

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def selector_fill(self, SelectorObject selector,
                      np.ndarray source,
                      np.ndarray dest = None,
                      np.int64_t offset = 0, int dims = 1,
                      int domain_id = -1):
        # This is where we use the selector to transplant from one to the
        # other.  Note that we *do* apply the selector here.
        cdef np.int64_t num_cells = -1
        cdef np.int64_t ind
        cdef np.int64_t sfc, sfci = -1
        cdef np.float64_t pos[3]
        cdef np.float64_t dpos[3]
        cdef int dim, status, filled = 0
        cdef int num_oct_levels, level
        cdef int max_level = self.artio_handle.max_level
        cdef int *num_octs_per_level = <int *>malloc(
            (max_level + 1)*sizeof(int))
        cdef char *sdata = <char*> source.data
        cdef char *ddata
        cdef int ss = source.dtype.itemsize
        cdef np.ndarray[np.uint8_t, ndim=1, cast=True] mask
        mask = self.mask(selector)
        if dest is None:
            # Note that RAMSES can have partial refinement inside an Oct.  This
            # means we actually do want the number of Octs, not the number of
            # cells.
            num_cells = self._last_mask_sum
            if dims > 1:
                dest = np.zeros((num_cells, dims), dtype=source.dtype,
                    order='C')
            else:
                dest = np.zeros(num_cells, dtype=source.dtype, order='C')
        ddata = (<char*>dest.data) + offset*ss*dims
        ind = 0
        for sfc in range(self.sfc_start, self.sfc_end + 1):
            if self.sfc_mask[sfc - self.sfc_start] == 0: continue
            sfci += 1
            if mask[sfci] == 0: continue
            memcpy(ddata, sdata + ind, dims * ss)
            ddata += dims * ss
            filled += 1
            ind += ss * dims
        if num_cells >= 0:
            return dest
        return filled

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def mask(self, SelectorObject selector, np.int64_t num_cells = -1,
             int domain_id = -1):
        # We take a domain_id here to avoid subclassing
        cdef int i
        cdef np.float64_t pos[3]
        cdef np.int64_t sfc, sfci = -1
        if self._last_selector_id == hash(selector):
            return self._last_mask
        cdef np.ndarray[np.uint8_t, ndim=1] mask
        mask = np.zeros((self.nsfc), dtype="uint8")
        self._last_mask_sum = 0
        for sfc in range(self.sfc_start, self.sfc_end + 1):
            if self.sfc_mask[sfc - self.sfc_start] == 0: continue
            sfci += 1
            self.sfc_to_pos(sfc, pos)
            if selector.select_cell(pos, self.dds) == 0: continue
            mask[sfci] = 1
            self._last_mask_sum += 1
        self._last_mask = mask.astype("bool")
        self._last_selector_id = hash(selector)
        return self._last_mask


    def fill_sfc_particles(self, fields):
        rv = read_sfc_particles(self.artio_handle,
                                self.sfc_start, self.sfc_end,
                                1, fields, self.range_handler.doct_count,
                                self.range_handler.pcount)
        return rv

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_sfc(self, SelectorObject selector, field_indices):
        cdef np.ndarray[np.float64_t, ndim=1] dest
        cdef int n, status, i, di, num_oct_levels, nf, ngv, max_level
        cdef np.int64_t sfc, num_cells, sfci = -1
        cdef np.float64_t val
        cdef double dpos[3]
        max_level = self.artio_handle.max_level
        cdef int *num_octs_per_level = <int *>malloc(
            (max_level + 1)*sizeof(int))
        # We duplicate some of the grid_variables stuff here so that we can
        # potentially release the GIL
        nf = len(field_indices)
        ngv = self.artio_handle.num_grid_variables
        cdef float *grid_variables = <float *>malloc(
            ngv * sizeof(float))
        cdef np.ndarray[np.uint8_t, ndim=1, cast=True] mask
        mask = self.mask(selector, -1)
        num_cells = self._last_mask_sum
        tr = []
        for i in range(nf):
            tr.append(np.zeros(num_cells, dtype="float64"))
        cdef int* field_ind = <int*> malloc(
            nf * sizeof(int))
        cdef np.float64_t **field_vals = <np.float64_t**> malloc(
            nf * sizeof(np.float64_t*))
        for i in range(nf):
            field_ind[i] = field_indices[i]
            # This zeros should be an empty once we handle the root grid
            dest = tr[i]
            field_vals[i] = <np.float64_t*> dest.data
        # First we need to walk the mesh in the file.  Then we fill in the dest
        # location based on the file index.
        cdef int filled = 0
        cdef float **mesh_data = self.range_handler.root_mesh_data
        if mesh_data == NULL:
            status = artio_grid_cache_sfc_range(self.handle, self.sfc_start,
                                                self.sfc_end)
            check_artio_status(status)
            for sfc in range(self.sfc_start, self.sfc_end + 1):
                if self.sfc_mask[sfc - self.sfc_start] == 0: continue
                sfci += 1
                if mask[sfci] == 0: continue
                status = artio_grid_read_root_cell_begin( self.handle,
                    sfc, dpos, grid_variables, &num_oct_levels,
                    num_octs_per_level)
                check_artio_status(status)
                for i in range(nf):
                    field_vals[i][filled] = grid_variables[field_ind[i]]
                filled += 1
                status = artio_grid_read_root_cell_end(self.handle)
                check_artio_status(status)
            status = artio_grid_clear_sfc_cache(self.handle)
            check_artio_status(status)
        else:
            for sfc in range(self.sfc_start, self.sfc_end + 1):
                if self.sfc_mask[sfc - self.sfc_start] == 0: continue
                sfci += 1
                if mask[sfci] == 0: continue
                for i in range(nf):
                    field_vals[i][filled] = mesh_data[field_ind[i]][
                        sfc - self.sfc_start]
                filled += 1
        # Now we have all our sources.
        free(field_ind)
        free(field_vals)
        free(grid_variables)
        free(num_octs_per_level)
        return tr

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def deposit(self, ParticleDepositOperation pdeposit,
                SelectorObject selector,
                np.ndarray[np.float64_t, ndim=2] positions,
                fields):
        # This implements the necessary calls to enable particle deposition to
        # occur as needed.
        cdef int nf, i, j
        cdef np.int64_t sfc, sfci
        if fields is None:
            fields = []
        nf = len(fields)
        cdef np.float64_t[::cython.view.indirect, ::1] field_pointers 
        if nf > 0: field_pointers = OnceIndirect(fields)
        cdef np.float64_t[:] field_vals = np.empty(nf, dtype="float64")
        cdef np.ndarray[np.uint8_t, ndim=1, cast=True] mask
        mask = self.mask(selector, -1)
        cdef np.ndarray[np.int64_t, ndim=1] domain_ind
        domain_ind = np.zeros(self.sfc_end - self.sfc_start + 1,
                              dtype="int64") - 1
        j = 0
        sfci = -1
        for sfc in range(self.sfc_start, self.sfc_end + 1):
            if self.sfc_mask[sfc - self.sfc_start] == 0: continue
            sfci += 1
            if mask[sfci] == 0:
                continue
            domain_ind[sfc - self.sfc_start] = j
            j += 1
        cdef np.float64_t pos[3]
        cdef np.float64_t left_edge[3]
        cdef int coords[3]
        cdef int dims[3]
        dims[0] = dims[1] = dims[2] = 1
        cdef np.int64_t offset, moff
        cdef np.int64_t numpart = positions.shape[0]
        for i in range(positions.shape[0]):
            for j in range(nf):
                field_vals[j] = field_pointers[j][i]
            for j in range(3):
                pos[j] = positions[i, j]
                coords[j] = <int>((pos[j] - self.DLE[j])/self.dds[j])
            sfc = artio_sfc_index(self.artio_handle.handle, coords)
            if sfc < self.sfc_start or sfc > self.sfc_end: continue
            offset = domain_ind[sfc - self.sfc_start]
            if offset < 0: continue
            # Check that we found the oct ...
            for j in range(3):
                left_edge[j] = coords[j] * self.dds[j] + self.DLE[j]
            pdeposit.process(dims, i, left_edge, self.dds,
                         offset, pos, field_vals, sfc)
            if pdeposit.update_values == 1:
                for j in range(nf):
                    field_pointers[j][i] = field_vals[j]

cdef class SFCRangeSelector(SelectorObject):

    cdef SelectorObject base_selector
    cdef ARTIOSFCRangeHandler range_handler
    cdef ARTIORootMeshContainer mesh_container
    cdef np.int64_t sfc_start, sfc_end

    def __init__(self, dobj):
        self.base_selector = dobj.base_selector
        self.mesh_container = dobj.oct_handler
        self.range_handler = self.mesh_container.range_handler
        self.sfc_start = self.mesh_container.sfc_start
        self.sfc_end = self.mesh_container.sfc_end

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def select_grids(self,
                     np.ndarray[np.float64_t, ndim=2] left_edges,
                     np.ndarray[np.float64_t, ndim=2] right_edges,
                     np.ndarray[np.int32_t, ndim=2] levels):
        raise RuntimeError

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
        return self.select_point(pos)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_point(self, np.float64_t pos[3]) nogil:
        cdef np.int64_t sfc = self.mesh_container.pos_to_sfc(pos)
        if sfc > self.sfc_end: return 0
        cdef np.int64_t oc = self.range_handler.doct_count[
            sfc - self.sfc_start]
        if oc > 0: return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        return self.base_selector.select_bbox(left_edge, right_edge)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) nogil:
        # Because visitors now use select_grid, we should be explicitly
        # checking this.
        return self.base_selector.select_grid(left_edge, right_edge, level, o)

    def _hash_vals(self):
        return (hash(self.base_selector), self.sfc_start, self.sfc_end)

sfc_subset_selector = AlwaysSelector
#sfc_subset_selector = SFCRangeSelector




""" 
API for yt.data_objects



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .grid_patch import \
    AMRGridPatch

from .octree_subset import \
    OctreeSubset

from .static_output import \
    Dataset

from .profiles import \
    create_profile, \
    Profile1D, \
    Profile2D, \
    Profile3D, \
    ParticleProfile

from .time_series import \
    DatasetSeries, \
    DatasetSeriesObject

from .analyzer_objects import \
    AnalysisTask, analysis_task

from .data_containers import \
    data_object_registry

from . import construction_data_containers as __cdc
from . import selection_data_containers as __sdc

from .image_array import \
    ImageArray

from .particle_filters import \
    particle_filter, \
    add_particle_filter



"""
AHF-specific IO functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from operator import attrgetter

import numpy as np

from yt.funcs import \
    mylog
from yt.utilities.exceptions import \
    YTDomainOverflow
from yt.utilities.io_handler import \
    BaseIOHandler
from yt.utilities.lib.geometry_utils import \
    compute_morton


class IOHandlerAHFHalos(BaseIOHandler):
    _particle_reader = False
    _dataset_type = 'ahf'

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_particle_coords(self, chunks, ptf):
        # This needs to *yield* a series of tuples of (ptype, (x, y, z)).
        # chunks is a list of chunks, and ptf is a dict where the keys are
        # ptypes and the values are lists of fields.
        for data_file in self._get_data_files(chunks, ptf):
            pos = data_file._get_particle_positions('halos')
            x, y, z = (pos[:, i] for i in range(3))
            yield 'halos', (x, y, z)

    def _read_particle_fields(self, chunks, ptf, selector):
        # This gets called after the arrays have been allocated.  It needs to
        # yield ((ptype, field), data) where data is the masked results of
        # reading ptype, field and applying the selector to the data read in.
        # Selector objects have a .select_points(x,y,z) that returns a mask, so
        # you need to do your masking here.
        for data_file in self._get_data_files(chunks, ptf):
            cols = []
            for field_list in ptf.values():
                cols.extend(field_list)
            cols = list(set(cols))
            halos = data_file.read_data(usecols=cols)
            pos = data_file._get_particle_positions('halos')
            x, y, z = (pos[:, i] for i in range(3))
            yield 'halos', (x, y, z)
            mask = selector.select_points(x, y, z, 0.0)
            del x, y, z
            if mask is None: continue
            for ptype, field_list in sorted(ptf.items()):
                for field in field_list:
                    data = halos[field][mask].astype('float64')
                    yield (ptype, field), data

    def _initialize_index(self, data_file, regions):
        halos = data_file.read_data(usecols=['ID'])
        pcount = len(halos['ID'])
        morton = np.empty(pcount, dtype='uint64')
        mylog.debug('Initializing index % 5i (% 7i particles)',
                    data_file.file_id, pcount)
        if pcount == 0:
            return morton
        ind = 0
        pos = data_file._get_particle_positions('halos')
        pos = data_file.ds.arr(pos, 'code_length')
        dle = self.ds.domain_left_edge
        dre = self.ds.domain_right_edge
        if np.any(pos.min(axis=0) < dle) or np.any(pos.max(axis=0) > dre):
            raise YTDomainOverflow(pos.min(axis=0),
                                   pos.max(axis=0),
                                   dle, dre)
        regions.add_data_file(pos, data_file.file_id)
        morton[ind:ind+pos.shape[0]] = compute_morton(
            pos[:, 0], pos[:, 1], pos[:, 2], dle, dre)
        return morton

    def _count_particles(self, data_file):
        halos = data_file.read_data(usecols=['ID'])
        return {'halos': len(halos['ID'])}

    def _identify_fields(self, data_file):
        fields = [('halos', f) for f in data_file.col_names]
        return fields, {}

    # Helper methods

    def _get_data_files(self, chunks, ptf):
        # Only support halo reading for now.
        assert len(ptf) == 1
        assert list(ptf.keys())[0] == 'halos'
        # Get data_files
        chunks = list(chunks)
        data_files = set([])
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        data_files = sorted(data_files, key=attrgetter('filename'))
        for data_file in data_files:
            yield data_file



"""
The default unit symbol lookup table.


"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.units import dimensions
from yt.utilities.physical_ratios import \
    cm_per_pc, cm_per_ly, cm_per_au, cm_per_rsun, cm_per_m, \
    mass_sun_grams, sec_per_year, sec_per_day, sec_per_hr, \
    sec_per_min, temp_sun_kelvin, luminosity_sun_ergs_per_sec, \
    metallicity_sun, erg_per_eV, amu_grams, mass_electron_grams, \
    cm_per_ang, jansky_cgs, mass_jupiter_grams, mass_earth_grams, \
    kelvin_per_rankine, speed_of_light_cm_per_s, planck_length_cm, \
    planck_charge_esu, planck_energy_erg, planck_mass_grams, \
    planck_temperature_K, planck_time_s, mass_hydrogen_grams, \
    grams_per_pound, standard_gravity_cm_per_s2, pascal_per_atm, \
    newton_cgs, cm_per_rearth, cm_per_rjup
import numpy as np

# Lookup a unit symbol with the symbol string, and provide a tuple with the
# conversion factor to cgs and dimensionality.

default_unit_symbol_lut = {
    # base
    "g":  (1.0, dimensions.mass, 0.0, r"\rm{g}"),
    "s":  (1.0, dimensions.time, 0.0, r"\rm{s}"),
    "K":  (1.0, dimensions.temperature, 0.0, r"\rm{K}"),
    "radian": (1.0, dimensions.angle, 0.0, r"\rm{radian}"),

    # other cgs
    "dyne": (1.0, dimensions.force, 0.0, r"\rm{dyn}"),
    "erg":  (1.0, dimensions.energy, 0.0, r"\rm{erg}"),
    "esu":  (1.0, dimensions.charge_cgs, 0.0, r"\rm{esu}"),
    "gauss": (1.0, dimensions.magnetic_field_cgs, 0.0, r"\rm{G}"),
    "degC": (1.0, dimensions.temperature, -273.15, r"^\circ\rm{C}"),
    "statA": (1.0, dimensions.current_cgs, 0.0, r"\rm{statA}"),
    "statV": (1.0, dimensions.electric_potential_cgs, 0.0, r"\rm{statV}"),
    "statohm": (1.0, dimensions.resistance_cgs, 0.0, r"\rm{statohm}"),

    # some SI
    "m": (1.0e2, dimensions.length, 0.0, r"\rm{m}"),
    "J": (1.0e7, dimensions.energy, 0.0, r"\rm{J}"),
    "W": (1.0e7, dimensions.power, 0.0, r"\rm{W}"),
    "Hz": (1.0, dimensions.rate, 0.0, r"\rm{Hz}"),
    "N": (1.0e5, dimensions.force, 0.0, r"\rm{N}"),
    "C": (1.0, dimensions.charge_mks, 0.0, r"\rm{C}"),
    "A": (1.0, dimensions.current_mks, 0.0, r"\rm{A}"),
    "T": (1000.0, dimensions.magnetic_field_mks, 0.0, r"\rm{T}"),
    "Pa": (10.0, dimensions.pressure, 0.0, r"\rm{Pa}"),
    "V": (1.0e7, dimensions.electric_potential_mks, 0.0, r"\rm{V}"),
    "ohm": (1.0e7, dimensions.resistance_mks, 0.0, r"\Omega"),

    # Imperial and other non-metric units
    "ft": (30.48, dimensions.length, 0.0, r"\rm{ft}"),
    "mile": (160934, dimensions.length, 0.0, r"\rm{mile}"),
    "degF": (kelvin_per_rankine, dimensions.temperature, -459.67,
             "^\circ\rm{F}"),
    "R": (kelvin_per_rankine, dimensions.temperature, 0.0, r"^\circ\rm{R}"),
    "lbf": (grams_per_pound*standard_gravity_cm_per_s2, dimensions.force, 0.0, r"\rm{lbf}"),
    "lbm": (grams_per_pound, dimensions.mass, 0.0, r"\rm{lbm}"),
    "atm": (pascal_per_atm*10., dimensions.pressure, 0.0, r"\rm{atm}"),

    # dimensionless stuff
    "h": (1.0, dimensions.dimensionless, 0.0, r"h"),  # needs to be added for rho_crit_now
    "dimensionless": (1.0, dimensions.dimensionless, 0.0, r""),

    # times
    "min": (sec_per_min, dimensions.time, 0.0, r"\rm{min}"),
    "hr":  (sec_per_hr, dimensions.time, 0.0, r"\rm{hr}"),
    "day": (sec_per_day, dimensions.time, 0.0, r"\rm{d}"),
    "d":   (sec_per_day, dimensions.time, 0.0, r"\rm{d}"),
    "yr":  (sec_per_year, dimensions.time, 0.0, r"\rm{yr}"),

    # Velocities
    "c": (speed_of_light_cm_per_s, dimensions.velocity, 0.0, r"\rm{c}"),

    # Solar units
    "Msun": (mass_sun_grams, dimensions.mass, 0.0, r"M_\odot"),
    "msun": (mass_sun_grams, dimensions.mass, 0.0, r"M_\odot"),
    "Rsun": (cm_per_rsun, dimensions.length, 0.0, r"R_\odot"),
    "rsun": (cm_per_rsun, dimensions.length, 0.0, r"R_\odot"),
    "R_sun": (cm_per_rsun, dimensions.length, 0.0, r"R_\odot"),
    "r_sun": (cm_per_rsun, dimensions.length, 0.0, r"R_\odot"),
    "Lsun": (luminosity_sun_ergs_per_sec, dimensions.power, 0.0, r"L_\odot"),
    "Tsun": (temp_sun_kelvin, dimensions.temperature, 0.0, r"T_\odot"),
    "Zsun": (metallicity_sun, dimensions.dimensionless, 0.0, r"Z_\odot"),
    "Mjup": (mass_jupiter_grams, dimensions.mass, 0.0, r"M_{\rm{Jup}}"),
    "Mearth": (mass_earth_grams, dimensions.mass, 0.0, r"M_\oplus"),

    # astro distances
    "AU": (cm_per_au, dimensions.length, 0.0, r"\rm{AU}"),
    "au": (cm_per_au, dimensions.length, 0.0, r"\rm{AU}"),
    "ly": (cm_per_ly, dimensions.length, 0.0, r"\rm{ly}"),
    "pc": (cm_per_pc, dimensions.length, 0.0, r"\rm{pc}"),

    # angles
    "degree": (np.pi/180., dimensions.angle, 0.0, r"\rm{deg}"),  # degrees
    "arcmin": (np.pi/10800., dimensions.angle, 0.0,
               r"\rm{arcmin}"),  # arcminutes
    "arcsec": (np.pi/648000., dimensions.angle, 0.0,
               r"\rm{arcsec}"),  # arcseconds
    "mas": (np.pi/648000000., dimensions.angle, 0.0,
            r"\rm{mas}"),  # milliarcseconds
    "hourangle": (np.pi/12., dimensions.angle, 0.0, r"\rm{HA}"),  # hour angle
    "steradian": (1.0, dimensions.solid_angle, 0.0, r"\rm{sr}"),
    "lat": (-np.pi/180.0, dimensions.angle, 90.0, r"\rm{Latitude}"),
    "lon": (np.pi/180.0, dimensions.angle, -180.0, r"\rm{Longitude}"),

    # misc
    "eV": (erg_per_eV, dimensions.energy, 0.0, r"\rm{eV}"),
    "amu": (amu_grams, dimensions.mass, 0.0, r"\rm{amu}"),
    "angstrom": (cm_per_ang, dimensions.length, 0.0, r"\AA"),
    "Jy": (jansky_cgs, dimensions.specific_flux, 0.0, r"\rm{Jy}"),
    "counts": (1.0, dimensions.dimensionless, 0.0, r"\rm{counts}"),
    "photons": (1.0, dimensions.dimensionless, 0.0, r"\rm{photons}"),
    "me": (mass_electron_grams, dimensions.mass, 0.0, r"m_e"),
    "mp": (mass_hydrogen_grams, dimensions.mass, 0.0, r"m_p"),
    "mol": (1.0/amu_grams, dimensions.dimensionless, 0.0, r"\rm{mol}"),
    'Sv': (cm_per_m**2, dimensions.specific_energy, 0.0, r"\rm{Sv}"),
    "rayleigh": (0.25e6/np.pi, dimensions.count_intensity, 0.0, r"\rm{R}"),

    # for AstroPy compatibility
    "solMass": (mass_sun_grams, dimensions.mass, 0.0, r"M_\odot"),
    "solRad": (cm_per_rsun, dimensions.length, 0.0, r"R_\odot"),
    "solLum": (luminosity_sun_ergs_per_sec, dimensions.power, 0.0, r"L_\odot"),
    "dyn": (1.0, dimensions.force, 0.0, r"\rm{dyn}"),
    "sr": (1.0, dimensions.solid_angle, 0.0, r"\rm{sr}"),
    "rad": (1.0, dimensions.angle, 0.0, r"\rm{rad}"),
    "deg": (np.pi/180., dimensions.angle, 0.0, r"\rm{deg}"),
    "Fr":  (1.0, dimensions.charge_cgs, 0.0, r"\rm{Fr}"),
    "G": (1.0, dimensions.magnetic_field_cgs, 0.0, r"\rm{G}"),
    "Angstrom": (cm_per_ang, dimensions.length, 0.0, r"\AA"),
    "statC": (1.0, dimensions.charge_cgs, 0.0, r"\rm{statC}"),

    # Planck units
    "m_pl": (planck_mass_grams, dimensions.mass, 0.0, r"m_{\rm{P}}"),
    "l_pl": (planck_length_cm, dimensions.length, 0.0, r"\ell_\rm{P}"),
    "t_pl": (planck_time_s, dimensions.time, 0.0, r"t_{\rm{P}}"),
    "T_pl": (planck_temperature_K, dimensions.temperature, 0.0, r"T_{\rm{P}}"),
    "q_pl": (planck_charge_esu, dimensions.charge_cgs, 0.0, r"q_{\rm{P}}"),
    "E_pl": (planck_energy_erg, dimensions.energy, 0.0, r"E_{\rm{P}}"),

    # Geometrized units
    "m_geom": (mass_sun_grams, dimensions.mass, 0.0, r"M_\odot"),
    "l_geom": (newton_cgs*mass_sun_grams/speed_of_light_cm_per_s**2, dimensions.length, 0.0, r"M_\odot"),
    "t_geom": (newton_cgs*mass_sun_grams/speed_of_light_cm_per_s**3, dimensions.time, 0.0, r"M_\odot"),

    # Some Solar System units
    "R_earth": (cm_per_rearth, dimensions.length, 0.0, r"R_\oplus"),
    "r_earth": (cm_per_rearth, dimensions.length, 0.0, r"R_\oplus"),
    "R_jup": (cm_per_rjup, dimensions.length, 0.0, r"R_\mathrm{Jup}"),
    "r_jup": (cm_per_rjup, dimensions.length, 0.0, r"R_\mathrm{Jup}"),
}

# This dictionary formatting from magnitude package, credit to Juan Reyero.
unit_prefixes = {
    'Y': 1e24,   # yotta
    'Z': 1e21,   # zetta
    'E': 1e18,   # exa
    'P': 1e15,   # peta
    'T': 1e12,   # tera
    'G': 1e9,    # giga
    'M': 1e6,    # mega
    'k': 1e3,    # kilo
    'd': 1e1,    # deci
    'c': 1e-2,   # centi
    'm': 1e-3,   # mili
    'u': 1e-6,   # micro
    'n': 1e-9,   # nano
    'p': 1e-12,  # pico
    'f': 1e-15,  # femto
    'a': 1e-18,  # atto
    'z': 1e-21,  # zepto
    'y': 1e-24,  # yocto
}

latex_prefixes = {
    "u": r"\mu",
    }

prefixable_units = [
    "m",
    "pc",
    "mcm",
    "pccm",
    "g",
    "eV",
    "s",
    "yr",
    "K",
    "dyne",
    "erg",
    "esu",
    "J",
    "Hz",
    "W",
    "gauss",
    "G",
    "Jy",
    "N",
    "T",
    "A",
    "C",
    "statA",
    "Pa",
    "V",
    "statV",
    "ohm",
    "statohm",
    "Sv",
]

default_base_units = {
    dimensions.mass: 'g',
    dimensions.length: 'cm',
    dimensions.time: 's',
    dimensions.temperature: 'K',
    dimensions.angle: 'radian',
    dimensions.current_mks: 'A',
}



#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.testing import \
    assert_allclose_units, \
    fake_amr_ds

def test_AM_value():
    ds = fake_amr_ds(fields=("Density", "velocity_x", "velocity_y", "velocity_z"), length_unit=0.5)

    sp = ds.sphere([.5]*3, (0.1, 'code_length'))

    x0 = sp.center
    v0 = ds.arr([1, 2, 3], 'km/s')

    sp.set_field_parameter('bulk_velocity', v0)

    X = (ds.arr([sp[k] for k in 'xyz']) - x0[:, None]).T
    V = (ds.arr([sp['velocity_'+k] for k in 'xyz']) - v0[:, None]).T

    sAM_manual = ds.arr(np.cross(X, V), X.units*V.units)
    sAM = ds.arr([sp['specific_angular_momentum_'+k] for k in 'xyz']).T

    assert_allclose_units(sAM_manual, sAM)



"""
This is a recursive function to return a depth-first octree



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
cimport numpy as np
cimport cython

cdef class position:
    cdef public int output_pos, refined_pos
    def __cinit__(self):
        self.output_pos = 0
        self.refined_pos = 0

cdef class OctreeGrid:
    cdef public object child_indices, fields, left_edges, dimensions, dx
    cdef public int level, offset
    def __cinit__(self,
                  np.ndarray[np.int32_t, ndim=3] child_indices,
                  np.ndarray[np.float64_t, ndim=4] fields,
                  np.ndarray[np.float64_t, ndim=1] left_edges,
                  np.ndarray[np.int32_t, ndim=1] dimensions,
                  np.ndarray[np.float64_t, ndim=1] dx,
                  int level, int offset):
        self.child_indices = child_indices
        self.fields = fields
        self.left_edges = left_edges
        self.dimensions = dimensions
        self.dx = dx
        self.level = level
        self.offset = offset

cdef class OctreeGridList:
    cdef public object grids
    def __cinit__(self, grids):
        self.grids = grids

    def __getitem__(self, int item):
        return self.grids[item]

@cython.boundscheck(False)
def RecurseOctreeDepthFirst(int i_i, int j_i, int k_i,
                            int i_f, int j_f, int k_f,
                            position curpos, int gi, 
                            np.ndarray[np.float64_t, ndim=2] output,
                            np.ndarray[np.int32_t, ndim=1] refined,
                            OctreeGridList grids):
    #cdef int s = curpos
    cdef int i, i_off, j, j_off, k, k_off, ci, fi
    cdef int child_i, child_j, child_k
    cdef OctreeGrid child_grid
    cdef OctreeGrid grid = grids[gi]
    cdef np.ndarray[np.float64_t, ndim=4] fields = grid.fields
    cdef np.ndarray[np.float64_t, ndim=1] leftedges = grid.left_edges
    cdef np.float64_t dx = grid.dx[0]
    cdef np.float64_t child_dx
    cdef np.ndarray[np.float64_t, ndim=1] child_leftedges
    cdef np.float64_t cx, cy, cz
    #here we go over the 8 octants
    #in general however, a mesh cell on this level
    #may have more than 8 children on the next level
    #so we find the int float center (cxyz) of each child cell
    # and from that find the child cell indices
    for i_off in range(i_f):
        i = i_off + i_i #index
        cx = (leftedges[0] + i*dx)
        for j_off in range(j_f):
            j = j_off + j_i
            cy = (leftedges[1] + j*dx)
            for k_off in range(k_f):
                k = k_off + k_i
                cz = (leftedges[2] + k*dx)
                ci = grid.child_indices[i,j,k]
                if ci == -1:
                    for fi in range(fields.shape[0]):
                        output[curpos.output_pos,fi] = fields[fi,i,j,k]
                    refined[curpos.refined_pos] = 0
                    curpos.output_pos += 1
                    curpos.refined_pos += 1
                else:
                    refined[curpos.refined_pos] = 1
                    curpos.refined_pos += 1
                    child_grid = grids[ci-grid.offset]
                    child_dx = child_grid.dx[0]
                    child_leftedges = child_grid.left_edges
                    child_i = int((cx - child_leftedges[0])/child_dx)
                    child_j = int((cy - child_leftedges[1])/child_dx)
                    child_k = int((cz - child_leftedges[2])/child_dx)
                    # s = Recurs.....
                    RecurseOctreeDepthFirst(child_i, child_j, child_k, 2, 2, 2,
                                        curpos, ci - grid.offset, output, refined, grids)

@cython.boundscheck(False)
def RecurseOctreeByLevels(int i_i, int j_i, int k_i,
                          int i_f, int j_f, int k_f,
                          np.ndarray[np.int32_t, ndim=1] curpos,
                          int gi, 
                          np.ndarray[np.float64_t, ndim=2] output,
                          np.ndarray[np.int32_t, ndim=2] genealogy,
                          np.ndarray[np.float64_t, ndim=2] corners,
                          OctreeGridList grids):
    cdef np.int32_t i, i_off, j, j_off, k, k_off, ci, fi
    cdef int child_i, child_j, child_k
    cdef OctreeGrid child_grid
    cdef OctreeGrid grid = grids[gi-1]
    cdef int level = grid.level
    cdef np.ndarray[np.int32_t, ndim=3] child_indices = grid.child_indices
    cdef np.ndarray[np.float64_t, ndim=4] fields = grid.fields
    cdef np.ndarray[np.float64_t, ndim=1] leftedges = grid.left_edges
    cdef np.float64_t dx = grid.dx[0]
    cdef np.float64_t child_dx
    cdef np.ndarray[np.float64_t, ndim=1] child_leftedges
    cdef np.float64_t cx, cy, cz
    cdef int cp
    s = None
    for i_off in range(i_f):
        i = i_off + i_i
        cx = (leftedges[0] + i*dx)
        for j_off in range(j_f):
            j = j_off + j_i
            cy = (leftedges[1] + j*dx)
            for k_off in range(k_f):
                k = k_off + k_i
                cz = (leftedges[2] + k*dx)
                cp = curpos[level]
                corners[cp, 0] = cx 
                corners[cp, 1] = cy 
                corners[cp, 2] = cz
                genealogy[curpos[level], 2] = level
                # always output data
                for fi in range(fields.shape[0]):
                    output[cp,fi] = fields[fi,i,j,k]
                ci = child_indices[i,j,k]
                if ci > -1:
                    child_grid = grids[ci-1]
                    child_dx = child_grid.dx[0]
                    child_leftedges = child_grid.left_edges
                    child_i = int((cx-child_leftedges[0])/child_dx)
                    child_j = int((cy-child_leftedges[1])/child_dx)
                    child_k = int((cz-child_leftedges[2])/child_dx)
                    # set current child id to id of next cell to examine
                    genealogy[cp, 0] = curpos[level+1] 
                    # set next parent id to id of current cell
                    genealogy[curpos[level+1]:curpos[level+1]+8, 1] = cp
                    s = RecurseOctreeByLevels(child_i, child_j, child_k, 2, 2, 2,
                                              curpos, ci, output, genealogy,
                                              corners, grids)
                curpos[level] += 1
    return s




"""
Gadget-specific fields




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.frontends.sph.fields import SPHFieldInfo
from yt.fields.particle_fields import add_volume_weighted_smoothed_field
from yt.utilities.physical_constants import mp, kb

class GadgetFieldInfo(SPHFieldInfo):

    def setup_particle_fields(self, ptype, *args, **kwargs):

        # setup some special fields that only make sense for SPH particles

        if (ptype, "FourMetalFractions") in self.ds.field_list:
            self._setup_four_metal_fractions(ptype)

        super(GadgetFieldInfo, self).setup_particle_fields(
            ptype, *args, **kwargs)

        if ptype in ("PartType0", "Gas"):
            self.setup_gas_particle_fields(ptype)

    def _setup_four_metal_fractions(self, ptype):
        """
        This function breaks the FourMetalFractions field (if present)
        into its four component metal fraction fields and adds
        corresponding metal density fields which will later get smoothed

        This gets used with the Gadget group0000 format
        as defined in the gadget_field_specs in frontends/gadget/definitions.py
        """
        metal_names = ['C', 'O', 'Si', 'Fe']
        for i, metal_name in enumerate(metal_names):

            # add the metal fraction fields
            def _Fraction_wrap(i):
                def _Fraction(field, data):
                    return data[(ptype, 'FourMetalFractions')][:,i]
                return _Fraction

            self.add_field( (ptype, metal_name+"_fraction"),
                            sampling_type="particle",
                            function=_Fraction_wrap(i),
                            units="")

            # add the metal density fields
            def _Density_wrap(i):
                def _Metal_density(field, data):
                    return data[(ptype, 'FourMetalFractions')][:,i] * \
                           data[(ptype, 'density')]
                return _Metal_density

            self.add_field( (ptype, metal_name+"_density"),
                            sampling_type="particle",
                            function=_Density_wrap(i),
                            units=self.ds.unit_system["density"])

    def setup_gas_particle_fields(self, ptype):
        if (ptype, "ElectronAbundance") in self.ds.field_list:
            def _temperature(field, data):
                # Assume cosmic abundances
                x_H = 0.76
                gamma = 5.0/3.0
                a_e = data[ptype, 'ElectronAbundance']
                mu = 4.0 / (3.0 * x_H + 1.0 + 4.0 * x_H * a_e)
                ret = data[ptype, "InternalEnergy"]*(gamma-1)*mu*mp/kb
                return ret.in_units(self.ds.unit_system["temperature"])
        else:
            def _temperature(field, data):
                # Assume cosmic abundances
                x_H = 0.76
                gamma = 5.0/3.0
                if data.has_field_parameter("mean_molecular_weight"):
                    mu = data.get_field_parameter("mean_molecular_weight")
                else:
                    # Assume zero ionization
                    mu = 4.0 / (3.0 * x_H + 1.0)
                ret = data[ptype, "InternalEnergy"]*(gamma-1)*mu*mp/kb
                return ret.in_units(self.ds.unit_system["temperature"])

        self.add_field(
            (ptype, "Temperature"),
            sampling_type="particle",
            function=_temperature,
            units=self.ds.unit_system["temperature"])
        self.alias((ptype, 'temperature'), (ptype, 'Temperature'))

        # For now, we hardcode num_neighbors.  We should make this configurable
        # in the future.
        num_neighbors = 64
        fn = add_volume_weighted_smoothed_field(
            ptype, "particle_position", "particle_mass", "smoothing_length",
            "density", "Temperature", self, num_neighbors)

        # Alias ("gas", "temperature") to the new smoothed Temperature field
        self.alias(("gas", "temperature"), fn[0])



"""
A means of running standalone commands with a shared set of options.



"""
from __future__ import print_function
from __future__ import absolute_import

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import argparse
import base64
import getpass
import numpy as np
import os
import sys
import subprocess
import tempfile
import json
import pprint
import textwrap

from yt.config import ytcfg, CURRENT_CONFIG_FILE
ytcfg["yt","__command_line"] = "True"
from yt.startup_tasks import parser, subparsers
from yt.funcs import \
    ensure_dir, \
    ensure_list, \
    get_hg_or_git_version, \
    get_yt_version, \
    mylog, \
    ensure_dir_exists, \
    update_hg_or_git, \
    enable_plugins, \
    download_file
from yt.extern.six import add_metaclass, string_types
from yt.extern.six.moves import urllib, input
from yt.extern.six.moves.urllib.parse import urlparse
from yt.extern.tqdm import tqdm
from yt.convenience import load
from yt.visualization.plot_window import \
    SlicePlot, \
    ProjectionPlot
from yt.utilities.metadata import get_metadata
from yt.utilities.configure import set_config
from yt.utilities.exceptions import \
    YTOutputNotIdentified, YTFieldNotParseable, YTCommandRequiresModule

# loading field plugins for backward compatibility, since this module
# used to do "from yt.mods import *"
if ytcfg.getboolean("yt","loadfieldplugins"):
    enable_plugins()

_default_colormap = ytcfg.get("yt", "default_colormap")

def _fix_ds(arg, *args, **kwargs):
    if os.path.isdir("%s" % arg) and \
        os.path.exists("%s/%s" % (arg,arg)):
        ds = load("%s/%s" % (arg,arg), *args, **kwargs)
    elif os.path.isdir("%s.dir" % arg) and \
        os.path.exists("%s.dir/%s" % (arg,arg)):
        ds = load("%s.dir/%s" % (arg,arg), *args, **kwargs)
    elif arg.endswith(".index"):
        ds = load(arg[:-10], *args, **kwargs)
    else:
        ds = load(arg, *args, **kwargs)
    return ds

def _add_arg(sc, arg):
    if isinstance(arg, string_types):
        arg = _common_options[arg].copy()
    argc = dict(arg.items())
    argnames = []
    if "short" in argc: argnames.append(argc.pop('short'))
    if "longname" in argc: argnames.append(argc.pop('longname'))
    sc.add_argument(*argnames, **argc)

def _print_failed_source_update(reinstall=False):
    print()
    print("The yt package is not installed from a git repository,")
    print("so you must update this installation manually.")
    if 'Continuum Analytics' in sys.version or 'Anaconda' in sys.version:
        # see http://stackoverflow.com/a/21318941/1382869 for why we need
        # to check both Continuum *and* Anaconda
        print()
        print("Since it looks like you are using a python installation")
        print("that is managed by conda, you may want to do:")
        print()
        print("    $ conda update yt")
        print()
        print("to update your yt installation.")
        if reinstall is True:
            print()
            print("To update all of your packages, you can do:")
            print()
            print("    $ conda update --all")
    else:
        print("If you manage your python dependencies with pip, you may")
        print("want to do:")
        print()
        print("    $ pip install -U yt")
        print()
        print("to update your yt installation.")

def _print_installation_information(path):
    import yt
    print()
    print("yt module located at:")
    print("    %s" % (path))
    if "YT_DEST" in os.environ:
        spath = os.path.join(
            os.environ["YT_DEST"], "src", "yt-supplemental")
        if os.path.isdir(spath):
            print("The supplemental repositories are located at:")
            print("    %s" % (spath))
    print()
    print("The current version of yt is:")
    print()
    print("---")
    print("Version = %s" % yt.__version__)
    vstring = get_hg_or_git_version(path)
    if vstring is not None:
        print("Changeset = %s" % vstring.strip())
    print("---")
    return vstring


def _get_girder_client():
    try:
        import girder_client
    except ImportError:
        raise YTCommandRequiresModule('girder_client')
    if not ytcfg.get("yt", "hub_api_key"):
        print("Before you can access the yt Hub you need an API key")
        print("In order to obtain one, either register by typing:")
        print("  yt hub register")
        print("or follow the instruction on:")
        print("  http://yt-project.org/docs/dev/sharing_data.html#obtaining-an-api-key")
        sys.exit()
    hub_url = urlparse(ytcfg.get("yt", "hub_url"))
    gc = girder_client.GirderClient(apiUrl=hub_url.geturl())
    gc.authenticate(apiKey=ytcfg.get("yt", "hub_api_key"))
    return gc


class FileStreamer:
    final_size = None
    next_sent = 0
    chunksize = 100*1024

    def __init__(self, f, final_size = None):
        location = f.tell()
        f.seek(0, os.SEEK_END)
        self.final_size = f.tell() - location
        f.seek(location)
        self.f = f

    def __iter__(self):
        with tqdm(total=self.final_size, desc='Uploading file',
                  unit='B', unit_scale=True) as pbar:
            while self.f.tell() < self.final_size:
                yield self.f.read(self.chunksize)
                pbar.update(self.chunksize)

_subparsers = {None: subparsers}
_subparsers_description = {
    'config': 'Get and set configuration values for yt',
    'hub': 'Interact with the yt Hub'
}
class YTCommandSubtype(type):
    def __init__(cls, name, b, d):
        type.__init__(cls, name, b, d)
        if cls.name is not None:
            names = ensure_list(cls.name)
            if cls.subparser not in _subparsers:
                try:
                    description = _subparsers_description[cls.subparser]
                except KeyError:
                    description = cls.subparser
                parent_parser = argparse.ArgumentParser(add_help=False)
                p = subparsers.add_parser(cls.subparser, help=description,
                                          description=description,
                                          parents=[parent_parser])
                _subparsers[cls.subparser] = p.add_subparsers(
                    title=cls.subparser, dest=cls.subparser)
            sp = _subparsers[cls.subparser]
            for name in names:
                sc = sp.add_parser(name, description=cls.description,
                                   help=cls.description)
                sc.set_defaults(func=cls.run)
                for arg in cls.args:
                    _add_arg(sc, arg)

@add_metaclass(YTCommandSubtype)
class YTCommand(object):
    args = ()
    name = None
    description = ""
    aliases = ()
    ndatasets = 1
    subparser = None

    @classmethod
    def run(cls, args):
        self = cls()
        # Check for some things we know; for instance, comma separated
        # field names should be parsed as tuples.
        if getattr(args, 'field', None) is not None and ',' in args.field:
            if args.field.count(",") > 1:
                raise YTFieldNotParseable(args.field)
            args.field = tuple(_.strip() for _ in args.field.split(","))
        if getattr(args, 'weight', None) is not None and ',' in args.weight:
            if args.weight.count(",") > 1:
                raise YTFieldNotParseable(args.weight)
            args.weight = tuple(_.strip() for _ in args.weight.split(","))
        # Some commands need to be run repeatedly on datasets
        # In fact, this is the rule and the opposite is the exception
        # BUT, we only want to parse the arguments once.
        if cls.ndatasets > 1:
            self(args)
        else:
            ds_args = getattr(args, "ds", [])
            if len(ds_args) > 1:
                datasets = args.ds
                for ds in datasets:
                    args.ds = ds
                    self(args)
            elif len(ds_args) == 0:
                datasets = []
                self(args)
            else:
                args.ds = getattr(args, 'ds', [None])[0]
                self(args)

class GetParameterFiles(argparse.Action):
    def __call__(self, parser, namespace, values, option_string = None):
        if len(values) == 1:
            datasets = values
        elif len(values) == 2 and namespace.basename is not None:
            datasets = ["%s%04i" % (namespace.basename, r)
                   for r in range(int(values[0]), int(values[1]), namespace.skip) ]
        else:
            datasets = values
        namespace.ds = [_fix_ds(ds) for ds in datasets]

_common_options = dict(
    all     = dict(longname="--all", dest="reinstall",
                   default=False, action="store_true",
                   help=("Reinstall the full yt stack in the current location."
                         "This option has been deprecated and will not have any"
                         "effect."),),
    ds      = dict(short="ds", action=GetParameterFiles,
                   nargs="+", help="datasets to run on"),
    ods     = dict(action=GetParameterFiles, dest="ds",
                   nargs="*", help="(Optional) datasets to run on"),
    axis    = dict(short="-a", longname="--axis",
                   action="store", type=int,
                   dest="axis", default=4,
                   help="Axis (4 for all three)"),
    log     = dict(short="-l", longname="--log",
                   action="store_true",
                   dest="takelog", default=True,
                   help="Use logarithmic scale for image"),
    linear  = dict(longname="--linear",
                   action="store_false",
                   dest="takelog",
                   help="Use linear scale for image"),
    text    = dict(short="-t", longname="--text",
                   action="store", type=str,
                   dest="text", default=None,
                   help="Textual annotation"),
    field   = dict(short="-f", longname="--field",
                   action="store", type=str,
                   dest="field", default="density",
                   help=("Field to color by, "
                         "use a comma to separate field tuple values")),
    weight  = dict(short="-g", longname="--weight",
                   action="store", type=str,
                   dest="weight", default=None,
                   help=("Field to weight projections with, "
                         "use a comma to separate field tuple values")),
    cmap    = dict(longname="--colormap",
                   action="store", type=str,
                   dest="cmap", default=_default_colormap,
                   help="Colormap name"),
    zlim    = dict(short="-z", longname="--zlim",
                   action="store", type=float,
                   dest="zlim", default=None,
                   nargs=2,
                   help="Color limits (min, max)"),
    dex     = dict(longname="--dex",
                   action="store", type=float,
                   dest="dex", default=None,
                   nargs=1,
                   help="Number of dex above min to display"),
    width   = dict(short="-w", longname="--width",
                   action="store", type=float,
                   dest="width", default=None,
                   help="Width in specified units"),
    unit    = dict(short="-u", longname="--unit",
                   action="store", type=str,
                   dest="unit", default='1',
                   help="Desired axes units"),
    center  = dict(short="-c", longname="--center",
                   action="store", type=float,
                   dest="center", default=None,
                   nargs=3,
                   help="Center, space separated (-1 -1 -1 for max)"),
    max     = dict(short="-m", longname="--max",
                   action="store_true",
                   dest="max",default=False,
                   help="Center the plot on the density maximum"),
    bn      = dict(short="-b", longname="--basename",
                   action="store", type=str,
                   dest="basename", default=None,
                   help="Basename of datasets"),
    output  = dict(short="-o", longname="--output",
                   action="store", type=str,
                   dest="output", default="frames/",
                   help="Folder in which to place output images"),
    outputfn= dict(short="-o", longname="--output",
                   action="store", type=str,
                   dest="output", default=None,
                   help="File in which to place output"),
    skip    = dict(short="-s", longname="--skip",
                   action="store", type=int,
                   dest="skip", default=1,
                   help="Skip factor for outputs"),
    proj    = dict(short="-p", longname="--projection",
                   action="store_true",
                   dest="projection", default=False,
                   help="Use a projection rather than a slice"),
    maxw    = dict(longname="--max-width",
                   action="store", type=float,
                   dest="max_width", default=1.0,
                   help="Maximum width in code units"),
    minw    = dict(longname="--min-width",
                   action="store", type=float,
                   dest="min_width", default=50,
                   help="Minimum width in units of smallest dx (default: 50)"),
    nframes = dict(short="-n", longname="--nframes",
                   action="store", type=int,
                   dest="nframes", default=100,
                   help="Number of frames to generate"),
    slabw   = dict(longname="--slab-width",
                   action="store", type=float,
                   dest="slab_width", default=1.0,
                   help="Slab width in specified units"),
    slabu   = dict(short="-g", longname="--slab-unit",
                   action="store", type=str,
                   dest="slab_unit", default='1',
                   help="Desired units for the slab"),
    ptype   = dict(longname="--particle-type",
                   action="store", type=int,
                   dest="ptype", default=2,
                   help="Particle type to select"),
    agecut  = dict(longname="--age-cut",
                   action="store", type=float,
                   dest="age_filter", default=None,
                   nargs=2,
                   help="Bounds for the field to select"),
    uboxes  = dict(longname="--unit-boxes",
                   action="store_true",
                   dest="unit_boxes",
                   help="Display heldsul unit boxes"),
    thresh  = dict(longname="--threshold",
                   action="store", type=float,
                   dest="threshold", default=None,
                   help="Density threshold"),
    dm_only = dict(longname="--all-particles",
                   action="store_false",
                   dest="dm_only", default=True,
                   help="Use all particles"),
    grids   = dict(longname="--show-grids",
                   action="store_true",
                   dest="grids", default=False,
                   help="Show the grid boundaries"),
    time    = dict(longname="--time",
                   action="store_true",
                   dest="time", default=False,
                   help="Print time in years on image"),
    contours    = dict(longname="--contours",
                   action="store",type=int,
                   dest="contours", default=None,
                   help="Number of Contours for Rendering"),
    contour_width  = dict(longname="--contour_width",
                   action="store",type=float,
                   dest="contour_width", default=None,
                   help="Width of gaussians used for rendering."),
    enhance   = dict(longname="--enhance",
                   action="store_true",
                   dest="enhance", default=False,
                   help="Enhance!"),
    valrange  = dict(short="-r", longname="--range",
                   action="store", type=float,
                   dest="valrange", default=None,
                   nargs=2,
                   help="Range, space separated"),
    up  = dict(longname="--up",
                   action="store", type=float,
                   dest="up", default=None,
                   nargs=3,
                   help="Up, space separated"),
    viewpoint  = dict(longname="--viewpoint",
                   action="store", type=float,
                   dest="viewpoint", default=[1., 1., 1.],
                   nargs=3,
                   help="Viewpoint, space separated"),
    pixels    = dict(longname="--pixels",
                   action="store",type=int,
                   dest="pixels", default=None,
                   help="Number of Pixels for Rendering"),
    halos   = dict(longname="--halos",
                   action="store", type=str,
                   dest="halos",default="multiple",
                   help="Run halo profiler on a 'single' halo or 'multiple' halos."),
    halo_radius = dict(longname="--halo_radius",
                       action="store", type=float,
                       dest="halo_radius",default=0.1,
                       help="Constant radius for profiling halos if using hop output files with no radius entry. Default: 0.1."),
    halo_radius_units = dict(longname="--halo_radius_units",
                             action="store", type=str,
                             dest="halo_radius_units",default="1",
                             help="Units for radius used with --halo_radius flag. Default: '1' (code units)."),
    halo_hop_style = dict(longname="--halo_hop_style",
                          action="store", type=str,
                          dest="halo_hop_style",default="new",
                          help="Style of hop output file.  'new' for yt_hop files and 'old' for enzo_hop files."),
    halo_dataset = dict(longname="--halo_dataset",
                               action="store", type=str,
                               dest="halo_dataset",default=None,
                               help="HaloProfiler dataset."),
    make_profiles = dict(longname="--make_profiles",
                         action="store_true", default=False,
                         help="Make profiles with halo profiler."),
    make_projections = dict(longname="--make_projections",
                            action="store_true", default=False,
                            help="Make projections with halo profiler.")

    )

# This code snippet is modified from Georg Brandl
def bb_apicall(endpoint, data, use_pass = True):
    uri = 'https://api.bitbucket.org/1.0/%s/' % endpoint
    # since bitbucket doesn't return the required WWW-Authenticate header when
    # making a request without Authorization, we cannot use the standard urllib2
    # auth handlers; we have to add the requisite header from the start
    if data is not None:
        data = urllib.parse.urlencode(data)
    req = urllib.request.Request(uri, data)
    if use_pass:
        username = input("Bitbucket Username? ")
        password = getpass.getpass()
        upw = '%s:%s' % (username, password)
        req.add_header('Authorization', 'Basic %s' % base64.b64encode(upw).strip())
    return urllib.request.urlopen(req).read()

class YTBugreportCmd(YTCommand):
    name = "bugreport"
    description = \
        """
        Report a bug in yt

        """

    def __call__(self, args):
        print("===============================================================")
        print()
        print("Hi there!  Welcome to the yt bugreport taker.")
        print()
        print("===============================================================")
        print("At any time in advance of the upload of the bug, you should feel free")
        print("to ctrl-C out and submit the bug report manually by going here:")
        print("   http://bitbucket.org/yt_analysis/yt/issues/new")
        print()
        print("Also, in order to submit a bug through this interface, you")
        print("need a Bitbucket account. If you don't have one, exit this ")
        print("bugreport now and run the 'yt bootstrap_dev' command to create one.")
        print()
        print("Have you checked the existing bug reports to make")
        print("sure your bug has not already been recorded by someone else?")
        print("   http://bitbucket.org/yt_analysis/yt/issues?status=new&status=open")
        print()
        print("Finally, are you sure that your bug is, in fact, a bug? It might")
        print("simply be a misunderstanding that could be cleared up by")
        print("visiting the yt irc channel or getting advice on the email list:")
        print("   http://yt-project.org/irc.html")
        print("   https://mail.python.org/archives/list/yt-users@python.org/")
        print()
        summary = input("Press <enter> if you remain firm in your conviction to continue.")
        print()
        print()
        print("Okay, sorry about that. How about a nice, pithy ( < 12 words )")
        print("summary of the bug?  (e.g. 'Particle overlay problem with parallel ")
        print("projections')")
        print()
        try:
            current_version = get_yt_version()
        except Exception:
            current_version = "Unavailable"
        summary = input("Summary? ")
        bugtype = "bug"
        data = dict(title = summary, type=bugtype)
        print()
        print("Okay, now let's get a bit more information.")
        print()
        print("Remember that if you want to submit a traceback, you can run")
        print("any script with --paste or --detailed-paste to submit it to")
        print("the pastebin and then include the link in this bugreport.")
        if "EDITOR" in os.environ:
            print()
            print("Press enter to spawn your editor, %s" % os.environ["EDITOR"])
            input()
            tf = tempfile.NamedTemporaryFile(delete=False)
            fn = tf.name
            tf.close()
            subprocess.call("$EDITOR %s" % fn, shell = True)
            content = open(fn).read()
            try:
                os.unlink(fn)
            except Exception: pass
        else:
            print()
            print("Couldn't find an $EDITOR variable.  So, let's just take")
            print("take input here.  Type up your summary until you're ready")
            print("to be done, and to signal you're done, type --- by itself")
            print("on a line to signal your completion.")
            print()
            print("(okay, type now)")
            print()
            lines = []
            while 1:
                line = input()
                if line.strip() == "---": break
                lines.append(line)
            content = "\n".join(lines)
        content = "Reporting Version: %s\n\n%s" % (current_version, content)
        endpoint = "repositories/yt_analysis/yt/issues"
        data['content'] = content
        print()
        print("===============================================================")
        print()
        print("Okay, we're going to submit with this:")
        print()
        print("Summary: %s" % (data['title']))
        print()
        print("---")
        print(content)
        print("---")
        print()
        print("===============================================================")
        print()
        print("Is that okay?  If not, hit ctrl-c.  Otherwise, enter means")
        print("'submit'.  Next we'll ask for your Bitbucket Username.")
        print("If you don't have one, run the 'yt bootstrap_dev' command.")
        print()
        input()
        retval = bb_apicall(endpoint, data, use_pass=True)
        import json
        retval = json.loads(retval)
        url = "http://bitbucket.org/yt_analysis/yt/issue/%s" % retval['local_id']
        print()
        print("===============================================================")
        print()
        print("Thanks for your bug report!  Together we'll make yt totally bug free!")
        print("You can view bug report here:")
        print("   %s" % url)
        print()
        print("Keep in touch!")
        print()


class YTHubRegisterCmd(YTCommand):
    subparser = "hub"
    name = "register"
    description = \
        """
        Register a user on the yt Hub: http://hub.yt/
        """
    def __call__(self, args):
        try:
            import requests
        except ImportError:
            raise YTCommandRequiresModule('requests')
        if ytcfg.get("yt", "hub_api_key") != "":
            print("You seem to already have an API key for the hub in")
            print("{} . Delete this if you want to force a".format(CURRENT_CONFIG_FILE))
            print("new user registration.")
            sys.exit()
        print("Awesome!  Let's start by registering a new user for you.")
        print("Here's the URL, for reference: http://hub.yt/ ")
        print()
        print("As always, bail out with Ctrl-C at any time.")
        print()
        print("What username would you like to go by?")
        print()
        username = input("Username? ")
        if len(username) == 0: sys.exit(1)
        print()
        print("To start out, what's your name?")
        print()
        first_name = input("First Name? ")
        if len(first_name) == 0: sys.exit(1)
        print()
        last_name = input("Last Name? ")
        if len(last_name) == 0: sys.exit(1)
        print()
        print("And your email address?")
        print()
        email = input("Email? ")
        if len(email) == 0: sys.exit(1)
        print()
        print("Please choose a password:")
        print()
        while 1:
            password1 = getpass.getpass("Password? ")
            password2 = getpass.getpass("Confirm? ")
            if len(password1) == 0: continue
            if password1 == password2: break
            print("Sorry, they didn't match!  Let's try again.")
            print()
        print()
        print("Okay, press enter to register.  You should receive a welcome")
        print("message at %s when this is complete." % email)
        print()
        input()

        data = dict(firstName=first_name, email=email, login=username,
                    password=password1, lastName=last_name, admin=False)
        hub_url = ytcfg.get("yt", "hub_url")
        req = requests.post(hub_url + "/user", data=data)

        if req.ok:
            headers = {'Girder-Token': req.json()['authToken']['token']}
        else:
            if req.status_code == 400:
                print("Registration failed with 'Bad request':")
                print(req.json()["message"])
            exit(1)
        print("User registration successful")
        print("Obtaining API key...")
        req = requests.post(hub_url + "/api_key", headers=headers,
                            data={'name': 'ytcmd', 'active': True})
        apiKey = req.json()["key"]

        print("Storing API key in configuration file")
        set_config("yt", "hub_api_key", apiKey)

        print()
        print("SUCCESS!")
        print()


class YTInstInfoCmd(YTCommand):
    name = ["instinfo", "version"]
    args = (
            dict(short="-u", longname="--update-source", action="store_true",
                 default = False,
                 help="Update the yt installation, if able"),
            dict(short="-o", longname="--output-version", action="store",
                  default = None, dest="outputfile",
                  help="File into which the current revision number will be" +
                       "stored")
           )
    description = \
        """
        Get some information about the yt installation

        """

    def __call__(self, opts):
        import pkg_resources
        yt_provider = pkg_resources.get_provider("yt")
        path = os.path.dirname(yt_provider.module_path)
        vstring = _print_installation_information(path)
        if vstring is not None:
            print("This installation CAN be automatically updated.")
            if opts.update_source:
                update_hg_or_git(path)
        elif opts.update_source:
            _print_failed_source_update()
        if vstring is not None and opts.outputfile is not None:
            open(opts.outputfile, "w").write(vstring)

class YTLoadCmd(YTCommand):
    name = "load"
    description = \
        """
        Load a single dataset into an IPython instance

        """

    args = ("ds", )

    def __call__(self, args):
        if args.ds is None:
            print("Could not load file.")
            sys.exit()
        import yt.mods
        import yt

        import IPython

        local_ns = yt.mods.__dict__.copy()
        local_ns['ds'] = args.ds
        local_ns['pf'] = args.ds
        local_ns['yt'] = yt

        try:
            from traitlets.config.loader import Config
        except ImportError:
            from IPython.config.loader import Config
        import sys
        cfg = Config()
        # prepend sys.path with current working directory
        sys.path.insert(0,'')
        IPython.embed(config=cfg,user_ns=local_ns)

class YTMapserverCmd(YTCommand):
    args = ("proj", "field", "weight", "linear", "center", "width", "cmap",
            dict(short="-a", longname="--axis", action="store", type=int,
                 dest="axis", default=0, help="Axis"),
            dict(short ="-o", longname="--host", action="store", type=str,
                 dest="host", default=None, help="IP Address to bind on"),
            dict(short='ds', nargs=1, type=str, help='The dataset to load.')
    )

    name = "mapserver"
    description = \
        """
        Serve a plot in a GMaps-style interface

        """

    def __call__(self, args):
        from yt.visualization.mapserver.pannable_map import PannableMapServer
        from yt.frontends.ramses.data_structures import RAMSESDataset

        # For RAMSES datasets, use the bbox feature to make the dataset load faster
        if RAMSESDataset._is_valid(args.ds) and args.center and args.width:
            kwa = dict(bbox= [[c - args.width/2 for c in args.center],
                              [c + args.width/2 for c in args.center]])
        else:
            kwa = dict()

        ds = _fix_ds(args.ds, **kwa)
        if args.center and args.width:
            center = args.center
            width = args.width
            ad = ds.box(left_edge = [c - args.width/2 for c in args.center],
                        right_edge = [c + args.width/2 for c in args.center])
        else:
            center = [0.5]*3
            width = 1.0
            ad = ds.all_data()

        if args.axis >= 4:
            print("Doesn't work with multiple axes!")
            return
        if args.projection:
            p = ProjectionPlot(ds, args.axis, args.field, weight_field=args.weight, data_source=ad,
                               center=center, width=width)
        else:
            p = SlicePlot(ds, args.axis, args.field, data_source=ad,
                               center=center, width=width)
        p.set_log('all', args.takelog)
        p.set_cmap('all', args.cmap)

        PannableMapServer(p.data_source, args.field, args.takelog, args.cmap)
        try:
            import bottle
        except ImportError:
            raise ImportError(
                "The mapserver functionality requires the bottle "
                "package to be installed. Please install using `pip "
                "install bottle`."
            )
        bottle.debug(True)
        if args.host is not None:
            colonpl = args.host.find(":")
            if colonpl >= 0:
                port = int(args.host.split(":")[-1])
                args.host = args.host[:colonpl]
            else:
                port = 8080
            bottle.run(server='auto', host=args.host, port=port)
        else:
            bottle.run(server='auto')


class YTPastebinCmd(YTCommand):
    name = "pastebin"
    args = (
             dict(short="-l", longname="--language", action="store",
                  default = None, dest="language",
                  help="Use syntax highlighter for the file in language"),
             dict(short="-L", longname="--languages", action="store_true",
                  default = False, dest="languages",
                  help="Retrieve a list of supported languages"),
             dict(short="-e", longname="--encoding", action="store",
                  default = 'utf-8', dest="encoding",
                  help="Specify the encoding of a file (default is "
                        "utf-8 or guessing if available)"),
             dict(short="-b", longname="--open-browser", action="store_true",
                  default = False, dest="open_browser",
                  help="Open the paste in a web browser"),
             dict(short="-p", longname="--private", action="store_true",
                  default = False, dest="private",
                  help="Paste as private"),
             dict(short="-c", longname="--clipboard", action="store_true",
                  default = False, dest="clipboard",
                  help="File to output to; else, print."),
             dict(short="file", type=str),
            )
    description = \
        """
        Post a script to an anonymous pastebin

        """

    def __call__(self, args):
        import yt.utilities.lodgeit as lo
        lo.main(args.file, languages=args.languages, language=args.language,
                 encoding=args.encoding, open_browser=args.open_browser,
                 private=args.private, clipboard=args.clipboard)

class YTPastebinGrabCmd(YTCommand):
    args = (dict(short="number", type=str),)
    name = "pastebin_grab"
    description = \
        """
        Print an online pastebin to STDOUT for local use.
        """

    def __call__(self, args):
        import yt.utilities.lodgeit as lo
        lo.main( None, download=args.number )

class YTHubStartNotebook(YTCommand):
    args = (
        dict(dest="folderId", default=ytcfg.get("yt", "hub_sandbox"),
             nargs="?",
             help="(Optional) Hub folder to mount inside the Notebook"),
    )
    description = \
        """
        Start the Jupyter Notebook on the yt Hub.
        """
    subparser = "hub"
    name = "start"
    def __call__(self, args):
        gc = _get_girder_client()

        # TODO: should happen server-side
        _id = gc._checkResourcePath(args.folderId)

        resp = gc.post("/notebook/{}".format(_id))
        try:
            print("Launched! Please visit this URL:")
            print("    https://tmpnb.hub.yt" + resp['url'])
            print()
        except (KeyError, TypeError):
            print("Something went wrong. The yt Hub responded with : ")
            print(resp)

class YTNotebookUploadCmd(YTCommand):
    args = (dict(short="file", type=str),)
    description = \
        """
        Upload an IPython Notebook to the yt Hub.
        """

    name = "upload_notebook"
    def __call__(self, args):
        gc = _get_girder_client()
        username = gc.get("/user/me")["login"]
        gc.upload(args.file, "/user/{}/Public".format(username))

        _id = gc.resourceLookup(
            "/user/{}/Public/{}".format(username, args.file))["_id"]
        _fid = next(gc.listFile(_id))["_id"]
        hub_url = urlparse(ytcfg.get("yt", "hub_url"))
        print("Upload successful!")
        print()
        print("To access your raw notebook go here:")
        print()
        print("  {}://{}/#item/{}".format(hub_url.scheme, hub_url.netloc, _id))
        print()
        print("To view your notebook go here:")
        print()
        print("  http://nbviewer.jupyter.org/urls/{}/file/{}/download".format(
            hub_url.netloc + hub_url.path, _fid))
        print()

class YTPlotCmd(YTCommand):
    args = ("width", "unit", "bn", "proj", "center", "zlim", "axis", "field",
            "weight", "skip", "cmap", "output", "grids", "time", "ds", "max",
            "log", "linear",
            dict(short="-fu", longname="--field-unit",
                 action="store", type=str,
                 dest="field_unit", default=None,
                 help="Desired field units"),
            dict(longname='--show-scale-bar',
                 action='store_true',
                 help="Annotate the plot with the scale"))

    name = "plot"

    description = \
        """
        Create a set of images

        """

    def __call__(self, args):
        ds = args.ds
        center = args.center
        if args.center == (-1,-1,-1):
            mylog.info("No center fed in; seeking.")
            v, center = ds.find_max("density")
        if args.max:
            v, center = ds.find_max("density")
        elif args.center is None:
            center = 0.5*(ds.domain_left_edge + ds.domain_right_edge)
        center = np.array(center)
        if ds.dimensionality < 3:
            dummy_dimensions = np.nonzero(ds.index.grids[0].ActiveDimensions <= 1)
            axes = ensure_list(dummy_dimensions[0][0])
        elif args.axis == 4:
            axes = range(3)
        else:
            axes = [args.axis]

        unit = args.unit
        if unit is None:
            unit = 'unitary'
        if args.width is None:
            width = None
        else:
            width = (args.width, args.unit)

        for ax in axes:
            mylog.info("Adding plot for axis %i", ax)
            if args.projection:
                plt = ProjectionPlot(ds, ax, args.field, center=center,
                                     width=width,
                                     weight_field=args.weight)
            else:
                plt = SlicePlot(ds, ax, args.field, center=center,
                                width=width)
            if args.grids:
                plt.annotate_grids()
            if args.time:
                plt.annotate_timestamp()
            if args.show_scale_bar:
                plt.annotate_scale()

            if args.field_unit:
                plt.set_unit(args.field, args.field_unit)

            plt.set_cmap(args.field, args.cmap)
            plt.set_log(args.field, args.takelog)
            if args.zlim:
                plt.set_zlim(args.field,*args.zlim)
            ensure_dir_exists(args.output)
            plt.save(os.path.join(args.output,"%s" % (ds)))


class YTRPDBCmd(YTCommand):
    name = "rpdb"
    description = \
        """
        Connect to a currently running (on localhost) rpd session.

        Commands run with --rpdb will trigger an rpdb session with any
        uncaught exceptions.

        """
    args = (
            dict(short="-t", longname="--task", action="store",
                 default = 0, dest='task',
                 help="Open a web browser."),
           )

    def __call__(self, args):
        from . import rpdb
        rpdb.run_rpdb(int(args.task))

class YTNotebookCmd(YTCommand):
    name = ["notebook"]
    args = (
            dict(short="-o", longname="--open-browser", action="store_true",
                 default = False, dest='open_browser',
                 help="Open a web browser."),
            dict(short="-p", longname="--port", action="store",
                 default = 0, dest='port',
                 help="Port to listen on; defaults to auto-detection."),
            dict(short="-prof", longname="--profile", action="store",
                 default = None, dest="profile",
                 help="The IPython profile to use when lauching the kernel."),
            dict(short="-n", longname="--no-password", action="store_true",
                 default = False, dest='no_password',
                 help="If set, do not prompt or use a password."),
            )
    description = \
        """
        Start the Jupyter Notebook locally.
        """
    def __call__(self, args):
        kwargs = {}
        from notebook.notebookapp import NotebookApp

        print("You must choose a password so that others cannot connect to "
              "your notebook.")
        pw = ytcfg.get("yt", "notebook_password")
        if len(pw) == 0 and not args.no_password:
            import IPython.lib
            pw = IPython.lib.passwd()
            print("If you would like to use this password in the future,")
            print("place a line like this inside the [yt] section in your")
            print("yt configuration file at ~/.config/yt/ytrc")
            print()
            print("notebook_password = %s" % pw)
            print()
        elif args.no_password:
            pw = None
        if args.port != 0:
            kwargs['port'] = int(args.port)
        if args.profile is not None:
            kwargs['profile'] = args.profile
        if pw is not None:
            kwargs['password'] = pw
        app = NotebookApp(open_browser=args.open_browser,
                          **kwargs)
        app.initialize(argv=[])
        print()
        print("***************************************************************")
        print()
        print("The notebook is now live at:")
        print()
        print("     http://127.0.0.1:%s/" % app.port)
        print()
        print("Recall you can create a new SSH tunnel dynamically by pressing")
        print("~C and then typing -L%s:localhost:%s" % (app.port, app.port))
        print("where the first number is the port on your local machine. ")
        print()
        print("If you are using %s on your machine already, try " \
              "-L8889:localhost:%s" % (app.port, app.port))
        print()
        print("***************************************************************")
        print()
        app.start()



class YTStatsCmd(YTCommand):
    args = ('outputfn','bn','skip','ds','field',
            dict(longname="--max", action='store_true', default=False,
                 dest='max', help="Display maximum of field requested through -f option."),
            dict(longname="--min", action='store_true', default=False,
                 dest='min', help="Display minimum of field requested through -f option."))
    name = "stats"
    description = \
        """
        Print stats and max/min value of a given field (if requested),
        for one or more datasets

        (default field is density)

        """

    def __call__(self, args):
        ds = args.ds
        ds.print_stats()
        vals = {}
        if args.field in ds.derived_field_list:
            if args.max is True:
                vals['min'] = ds.find_max(args.field)
                print("Maximum %s: %0.5e at %s" % (args.field,
                    vals['min'][0], vals['min'][1]))
            if args.min is True:
                vals['max'] = ds.find_min(args.field)
                print("Minimum %s: %0.5e at %s" % (args.field,
                    vals['max'][0], vals['max'][1]))
        if args.output is not None:
            t = ds.current_time * ds['years']
            with open(args.output, "a") as f:
                f.write("%s (%0.5e years)\n" % (ds, t))
                if 'min' in vals:
                    f.write('Minimum %s is %0.5e at %s\n' % (
                        args.field, vals['min'][0], vals['min'][1]))
                if 'max' in vals:
                    f.write('Maximum %s is %0.5e at %s\n' % (
                        args.field, vals['max'][0], vals['max'][1]))

class YTUpdateCmd(YTCommand):
    args = ("all", )
    name = "update"
    description = \
        """
        Update the yt installation to the most recent version

        """

    def __call__(self, opts):
        import pkg_resources
        yt_provider = pkg_resources.get_provider("yt")
        path = os.path.dirname(yt_provider.module_path)
        vstring = _print_installation_information(path)
        if vstring is not None:
            print()
            print("This installation CAN be automatically updated.")
            update_hg_or_git(path)
        else:
            _print_failed_source_update(opts.reinstall)


class YTDeleteImageCmd(YTCommand):
    args = (dict(short="delete_hash", type=str),)
    description = \
        """
        Delete image from imgur.com.

        """
    name = "delete_image"
    def __call__(self, args):
        headers = {'Authorization':
            'Client-ID {}'.format(ytcfg.get("yt", "imagebin_api_key"))}

        delete_url = ytcfg.get("yt", "imagebin_delete_url")
        req = urllib.request.Request(
            delete_url.format(delete_hash=args.delete_hash),
            headers=headers, method='DELETE')
        try:
            response = urllib.request.urlopen(req).read().decode()
        except urllib.error.HTTPError as e:
            print("ERROR", e)
            return {'deleted': False}

        rv = json.loads(response)
        if 'success' in rv and rv["success"]:
            print("\nImage successfully deleted!\n")
        else:
            print()
            print("Something has gone wrong!  Here is the server response:")
            print()
            pprint.pprint(rv)


class YTUploadImageCmd(YTCommand):
    args = (dict(short="file", type=str),)
    description = \
        """
        Upload an image to imgur.com.  Must be PNG.

        """
    name = "upload_image"
    def __call__(self, args):
        filename = args.file
        if not filename.endswith(".png"):
            print("File must be a PNG file!")
            return 1
        headers = {'Authorization':
            'Client-ID {}'.format(ytcfg.get("yt", "imagebin_api_key"))}

        image_data = base64.b64encode(open(filename, 'rb').read())
        parameters = {'image': image_data, type: 'base64',
                      'name': filename,
                      'title': "%s uploaded by yt" % filename}
        data = urllib.parse.urlencode(parameters).encode('utf-8')
        req = urllib.request.Request(
            ytcfg.get("yt", "imagebin_upload_url"), data=data, headers=headers)
        try:
            response = urllib.request.urlopen(req).read().decode()
        except urllib.error.HTTPError as e:
            print("ERROR", e)
            return {'uploaded':False}
        rv = json.loads(response)
        if 'data' in rv and 'link' in rv['data']:
            print()
            print("Image successfully uploaded!  You can find it at:")
            print("    %s" % (rv['data']['link']))
            print()
            print("If you'd like to delete it, use the following")
            print("    yt delete_image %s" % rv['data']['deletehash'])
            print()
        else:
            print()
            print("Something has gone wrong!  Here is the server response:")
            print()
            pprint.pprint(rv)


class YTUploadFileCmd(YTCommand):
    args = (dict(short="file", type=str),)
    description = \
        """
        Upload a file to yt's curldrop.

        """
    name = "upload"

    def __call__(self, args):
        try:
            import requests
        except ImportError:
            raise YTCommandRequiresModule('requests')

        fs = iter(FileStreamer(open(args.file, 'rb')))
        upload_url = ytcfg.get("yt", "curldrop_upload_url")
        r = requests.put(upload_url + "/" + os.path.basename(args.file),
                         data=fs)
        print()
        print(r.text)


class YTConfigGetCmd(YTCommand):
    subparser = 'config'
    name = 'get'
    description = 'get a config value'
    args = (dict(short='section', help='The section containing the option.'),
            dict(short='option', help='The option to retrieve.'))
    def __call__(self, args):
        from yt.utilities.configure import get_config
        print(get_config(args.section, args.option))


class YTConfigSetCmd(YTCommand):
    subparser = 'config'
    name = 'set'
    description = 'set a config value'
    args = (dict(short='section', help='The section containing the option.'),
            dict(short='option', help='The option to set.'),
            dict(short='value', help='The value to set the option to.'))
    def __call__(self, args):
        from yt.utilities.configure import set_config
        set_config(args.section, args.option, args.value)


class YTConfigRemoveCmd(YTCommand):
    subparser = 'config'
    name = 'rm'
    description = 'remove a config option'
    args = (dict(short='section', help='The section containing the option.'),
            dict(short='option', help='The option to remove.'))
    def __call__(self, args):
        from yt.utilities.configure import rm_config
        rm_config(args.section, args.option)


class YTConfigListCmd(YTCommand):
    subparser = 'config'
    name = 'list'
    description = 'show the config content'
    args = ()
    def __call__(self, args):
        from yt.utilities.configure import write_config
        write_config(sys.stdout)


class YTConfigMigrateCmd(YTCommand):
    subparser = 'config'
    name = 'migrate'
    description = 'migrate old config file'
    args = ()
    def __call__(self, args):
        from yt.utilities.configure import migrate_config
        migrate_config()


class YTSearchCmd(YTCommand):
    args = (dict(short="-o", longname="--output",
                 action="store", type=str,
                 dest="output", default="yt_index.json",
                 help="File in which to place output"),
            dict(longname="--check-all", short="-a",
                 help="Attempt to load every file",
                 action="store_true", default=False,
                 dest="check_all"),
            dict(longname="--full", short="-f",
                 help="Output full contents of parameter file",
                 action="store_true", default=False,
                 dest="full_output"),
            )
    description = \
        """
        Attempt to find outputs that yt can recognize in directories.
        """
    name = "search"
    def __call__(self, args):
        from yt.utilities.parameter_file_storage import \
            output_type_registry
        candidates = []
        for base, dirs, files in os.walk(".", followlinks=True):
            print("(% 10i candidates) Examining %s" % (len(candidates), base))
            recurse = []
            if args.check_all:
                candidates.extend([os.path.join(base, _) for _ in files])
            for _, otr in sorted(output_type_registry.items()):
                c, r = otr._guess_candidates(base, dirs, files)
                candidates.extend([os.path.join(base, _) for _ in c])
                recurse.append(r)
            if len(recurse) > 0 and not all(recurse):
                del dirs[:]
        # Now we have a ton of candidates.  We're going to do something crazy
        # and try to load each one.
        records = []
        for i, c in enumerate(sorted(candidates)):
            print("(% 10i/% 10i) Evaluating %s" % (i, len(candidates), c))
            try:
                record = get_metadata(c, args.full_output)
            except YTOutputNotIdentified:
                continue
            records.append(record)
        with open(args.output, "w") as f:
            json.dump(records, f, indent=4)
        print("Identified %s records output to %s" % (
              len(records), args.output))

class YTDownloadData(YTCommand):

    args = (
        dict(short="filename", action="store", type=str,
             help="The name of the file to download", nargs='?',
             default=''),
        dict(short="location", action="store", type=str, nargs='?',
             help="The location in which to place the file, can be "
                  "\"supp_data_dir\", \"test_data_dir\", or any valid "
                  "path on disk. ", default=''),
        dict(longname="--overwrite", short="-c",
             help="Overwrite existing file.",
             action="store_true", default=False),
        dict(longname="--list", short="-l",
             help="Display all available files.",
             action="store_true", default=False),
    )
    description = \
        """
        Download a file from http://yt-project.org/data and save it to a
        particular location. Files can be saved to the locations provided
        by the "test_data_dir" or "supp_data_dir" configuration entries, or
        any valid path to a location on disk.
        """
    name = "download"

    def __call__(self, args):
        if args.list:
            self.get_list()
            return
        if not args.filename:
            raise RuntimeError('You need to provide a filename. See --help '
                               'for details or use --list to get available '
                               'datasets.')
        elif not args.location:
            raise RuntimeError('You need to specify download location. See '
                               '--help for details.')
        data_url = "http://yt-project.org/data/%s" % args.filename
        if args.location in ["test_data_dir", "supp_data_dir"]:
            data_dir = ytcfg.get("yt", args.location)
            if data_dir == "/does/not/exist":
                raise RuntimeError("'%s' is not configured!" % args.location)
        else:
            data_dir = args.location
        if not os.path.exists(data_dir):
            print("The directory '%s' does not exist. Creating..." % data_dir)
            ensure_dir(data_dir)
        data_file = os.path.join(data_dir, args.filename)
        if os.path.exists(data_file) and not args.overwrite:
            raise IOError("File '%s' exists and overwrite=False!" % data_file)
        print("Attempting to download file: %s" % args.filename)
        fn = download_file(data_url, data_file)

        if not os.path.exists(fn):
            raise IOError("The file '%s' did not download!!" % args.filename)
        print("File: %s downloaded successfully to %s" %
              (args.filename, data_file))

    def get_list(self):
        data = urllib.request.urlopen(
            'http://yt-project.org/data/datafiles.json').read().decode('utf8')
        data = json.loads(data)
        for key in data:
            for ds in data[key]:
                ds['fullname'] = ds['url'].replace(
                    'http://yt-project.org/data/', '')
                print('{fullname} ({size}) type: {code}'.format(**ds))
                for line in textwrap.wrap(ds['description']):
                    print('\t', line)

def run_main():
    args = parser.parse_args()
    # The following is a workaround for a nasty Python 3 bug:
    # http://bugs.python.org/issue16308
    # http://bugs.python.org/issue9253
    try:
        getattr(args, "func")
    except AttributeError:
        parser.print_help()
        sys.exit(0)

    args.func(args)

if __name__ == "__main__": run_main()
